#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from unittest import TestCase

from holidays.countries.bonaire_sint_eustatius_and_saba import BonaireSintEustatiusAndSaba
from tests.common import CommonCountryTests


class TestBonaireSintEustatiusAndSaba(CommonCountryTests, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass(BonaireSintEustatiusAndSaba)

    def test_new_years_day(self):
        self.assertHolidayName("Nieuwjaarsdag", (f"{year}-01-01" for year in self.full_range))

    def test_good_friday(self):
        name = "Goede Vrijdag"
        self.assertHolidayName(
            name,
            "2020-04-10",
            "2021-04-02",
            "2022-04-15",
            "2023-04-07",
            "2024-03-29",
            "2025-04-18",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_sunday(self):
        name = "Eerste Paasdag"
        self.assertHolidayName(
            name,
            "2020-04-12",
            "2021-04-04",
            "2022-04-17",
            "2023-04-09",
            "2024-03-31",
            "2025-04-20",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_monday(self):
        name = "Tweede Paasdag"
        self.assertHolidayName(
            name,
            "2020-04-13",
            "2021-04-05",
            "2022-04-18",
            "2023-04-10",
            "2024-04-01",
            "2025-04-21",
        )
        self.assertHolidayName(name, self.full_range)

    def test_kings_day(self):
        name_queen = "Koninginnedag"
        name_king = "Koningsdag"
        self.assertHolidayName(
            name_queen, (f"{year}-04-30" for year in range(self.start_year, 2014))
        )
        self.assertNoHolidayName(name_queen, range(2014, self.end_year))
        self.assertHolidayName(
            name_king,
            "2014-04-26",
            "2015-04-27",
            "2020-04-27",
            "2021-04-27",
            "2022-04-27",
            "2023-04-27",
            "2024-04-27",
            "2025-04-26",
        )
        self.assertNoHoliday(
            "2014-04-27",
            "2025-04-27",
        )
        self.assertHolidayName(name_king, range(2014, self.end_year))
        self.assertNoHolidayName(name_king, range(self.start_year, 2014))

    def test_labor_day(self):
        self.assertHolidayName("Dag van de Arbeid", (f"{year}-05-01" for year in self.full_range))

    def test_ascension_day(self):
        name = "Hemelvaartsdag"
        self.assertHolidayName(
            name,
            "2020-05-21",
            "2021-05-13",
            "2022-05-26",
            "2023-05-18",
            "2024-05-09",
            "2025-05-29",
        )
        self.assertHolidayName(name, self.full_range)

    def test_whit_sunday(self):
        name = "Eerste Pinksterdag"
        self.assertHolidayName(
            name,
            "2020-05-31",
            "2021-05-23",
            "2022-06-05",
            "2023-05-28",
            "2024-05-19",
            "2025-06-08",
        )
        self.assertHolidayName(name, self.full_range)

    def test_christmas_day(self):
        self.assertHolidayName("Eerste Kerstdag", (f"{year}-12-25" for year in self.full_range))

    def test_boxing_day(self):
        self.assertHolidayName("Tweede Kerstdag", (f"{year}-12-26" for year in self.full_range))

    def test_rincon_day(self):
        name = "Rincondag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "BON":
                self.assertHolidayName(
                    name, holidays, (f"{year}-04-30" for year in range(2020, self.end_year))
                )
                self.assertNoHolidayName(name, holidays, range(self.start_year, 2020))
            else:
                self.assertNoHolidayName(name, holidays)

    def test_bonaire_day(self):
        name = "Bonairedag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "BON":
                self.assertHolidayName(
                    name, holidays, (f"{year}-09-06" for year in self.full_range)
                )
            else:
                self.assertNoHolidayName(name, holidays)

    def test_carnival_monday(self):
        name = "Dag na de carnavalsoptocht"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "SAB":
                self.assertHolidayName(
                    name,
                    holidays,
                    "2020-07-27",
                    "2021-07-26",
                    "2022-07-25",
                    "2023-07-31",
                    "2024-07-29",
                    "2025-07-28",
                )
                self.assertHolidayName(name, holidays, self.full_range)
            else:
                self.assertNoHolidayName(name, holidays)

    def test_saba_day(self):
        name = "Sabadag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "SAB":
                self.assertHolidayName(
                    name,
                    holidays,
                    "2020-12-04",
                    "2021-12-03",
                    "2022-12-02",
                    "2023-12-01",
                    "2024-12-06",
                    "2025-12-05",
                )
                self.assertHolidayName(name, holidays, self.full_range)
            else:
                self.assertNoHolidayName(name, holidays)

    def test_emancipation_day(self):
        name = "Emancipatiedag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "STA":
                self.assertHolidayName(
                    name, holidays, (f"{year}-07-01" for year in range(2022, self.end_year))
                )
            else:
                self.assertNoHolidayName(name, holidays)

    def test_statia_day(self):
        name = "Statiadag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "STA":
                self.assertHolidayName(
                    name, holidays, (f"{year}-11-16" for year in self.full_range)
                )
            else:
                self.assertNoHolidayName(name, holidays)

    def test_special_holiday(self):
        name = "Brugdag"
        self.assertNoHolidayName(name)

        for subdiv, holidays in self.subdiv_holidays.items():
            if subdiv == "BON":
                self.assertHolidayName(name, holidays, "2025-05-02")
            else:
                self.assertNoHolidayName(name, holidays)

    def test_l10n_default(self):
        self.assertLocalizedHolidays(
            ("2023-01-01", "Nieuwjaarsdag"),
            ("2023-04-07", "Goede Vrijdag"),
            ("2023-04-09", "Eerste Paasdag"),
            ("2023-04-10", "Tweede Paasdag"),
            ("2023-04-27", "Koningsdag"),
            ("2023-04-30", "Rincondag"),
            ("2023-05-01", "Dag van de Arbeid"),
            ("2023-05-18", "Hemelvaartsdag"),
            ("2023-05-28", "Eerste Pinksterdag"),
            ("2023-07-01", "Emancipatiedag"),
            ("2023-07-31", "Dag na de carnavalsoptocht"),
            ("2023-09-06", "Bonairedag"),
            ("2023-11-16", "Statiadag"),
            ("2023-12-01", "Sabadag"),
            ("2023-12-25", "Eerste Kerstdag"),
            ("2023-12-26", "Tweede Kerstdag"),
        )

    def test_l10n_en_bq(self):
        self.assertLocalizedHolidays(
            "en_BQ",
            ("2023-01-01", "New Year's Day"),
            ("2023-04-07", "Good Friday"),
            ("2023-04-09", "Easter Sunday"),
            ("2023-04-10", "Easter Monday"),
            ("2023-04-27", "King's Day"),
            ("2023-04-30", "Rincon Day"),
            ("2023-05-01", "Labour Day"),
            ("2023-05-18", "Ascension Day"),
            ("2023-05-28", "Whit Sunday"),
            ("2023-07-01", "Emancipation Day"),
            ("2023-07-31", "Carnival Monday"),
            ("2023-09-06", "Bonaire Day"),
            ("2023-11-16", "Statia Day"),
            ("2023-12-01", "Saba Day"),
            ("2023-12-25", "Christmas Day"),
            ("2023-12-26", "Second Day of Christmas"),
        )

    def test_l10n_en_us(self):
        self.assertLocalizedHolidays(
            "en_US",
            ("2023-01-01", "New Year's Day"),
            ("2023-04-07", "Good Friday"),
            ("2023-04-09", "Easter Sunday"),
            ("2023-04-10", "Easter Monday"),
            ("2023-04-27", "King's Day"),
            ("2023-04-30", "Rincon Day"),
            ("2023-05-01", "Labor Day"),
            ("2023-05-18", "Ascension Day"),
            ("2023-05-28", "Whit Sunday"),
            ("2023-07-01", "Emancipation Day"),
            ("2023-07-31", "Carnival Monday"),
            ("2023-09-06", "Bonaire Day"),
            ("2023-11-16", "Statia Day"),
            ("2023-12-01", "Saba Day"),
            ("2023-12-25", "Christmas Day"),
            ("2023-12-26", "Second Day of Christmas"),
        )

    def test_l10n_pap_bq(self):
        self.assertLocalizedHolidays(
            "pap_BQ",
            ("2023-01-01", "Aña nobo"),
            ("2023-04-07", "Bièrnèsantu"),
            ("2023-04-09", "Pasku di Resurekshon"),
            ("2023-04-10", "Di dos dia di Pasku di Resurekshon"),
            ("2023-04-27", "Dia di Rei"),
            ("2023-04-30", "Dia di Rincon"),
            ("2023-05-01", "Dia di labor"),
            ("2023-05-18", "Asenshon"),
            ("2023-05-28", "Dia di Pentekòste"),
            ("2023-07-01", "Dia di Emansipashon"),
            ("2023-07-31", "Djaluna di Carnaval"),
            ("2023-09-06", "Dia di Boneiru"),
            ("2023-11-16", "Dia di Statia"),
            ("2023-12-01", "Dia di Saba"),
            ("2023-12-25", "Pasku"),
            ("2023-12-26", "Di dos dia di Pasku"),
        )



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from unittest import TestCase

from holidays.countries.faroe_islands import FaroeIslands
from tests.common import CommonCountryTests


class TestFaroeIslands(CommonCountryTests, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass(FaroeIslands)

    def test_new_years_day(self):
        self.assertHolidayName("Nýggjársdagur", (f"{year}-01-01" for year in self.full_range))

    def test_maundy_thursday(self):
        name = "Skírhósdagur"
        self.assertHolidayName(
            name,
            "2020-04-09",
            "2021-04-01",
            "2022-04-14",
            "2023-04-06",
            "2024-03-28",
            "2025-04-17",
        )
        self.assertHolidayName(name, self.full_range)

    def test_good_friday(self):
        name = "Langifríggjadagur"
        self.assertHolidayName(
            name,
            "2020-04-10",
            "2021-04-02",
            "2022-04-15",
            "2023-04-07",
            "2024-03-29",
            "2025-04-18",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_sunday(self):
        name = "Páskadagur"
        self.assertHolidayName(
            name,
            "2020-04-12",
            "2021-04-04",
            "2022-04-17",
            "2023-04-09",
            "2024-03-31",
            "2025-04-20",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_monday(self):
        name = "Annar páskadagur"
        self.assertHolidayName(
            name,
            "2020-04-13",
            "2021-04-05",
            "2022-04-18",
            "2023-04-10",
            "2024-04-01",
            "2025-04-21",
        )
        self.assertHolidayName(name, self.full_range)

    def test_great_prayer_day(self):
        name = "Dýri biðidagur"
        self.assertHolidayName(
            name,
            "2020-05-08",
            "2021-04-30",
            "2022-05-13",
            "2023-05-05",
            "2024-04-26",
            "2025-05-16",
        )
        self.assertHolidayName(name, self.full_range)

    def test_ascension_day(self):
        name = "Kristi himmalsferðardagur"
        self.assertHolidayName(
            name,
            "2020-05-21",
            "2021-05-13",
            "2022-05-26",
            "2023-05-18",
            "2024-05-09",
            "2025-05-29",
        )
        self.assertHolidayName(name, self.full_range)

    def test_whit_sunday(self):
        name = "Hvítusunnudagur"
        self.assertHolidayName(
            name,
            "2020-05-31",
            "2021-05-23",
            "2022-06-05",
            "2023-05-28",
            "2024-05-19",
            "2025-06-08",
        )
        self.assertHolidayName(name, self.full_range)

    def test_whit_monday(self):
        name = "Annar hvítusunnudagur"
        self.assertHolidayName(
            name,
            "2020-06-01",
            "2021-05-24",
            "2022-06-06",
            "2023-05-29",
            "2024-05-20",
            "2025-06-09",
        )
        self.assertHolidayName(name, self.full_range)

    def test_saint_olafs_day(self):
        self.assertHolidayName("Ólavsøkudagur", (f"{year}-07-29" for year in self.full_range))

    def test_christmas_eve(self):
        self.assertHolidayName("Jólaaftan", (f"{year}-12-24" for year in self.full_range))

    def test_christmas_day(self):
        self.assertHolidayName("Jóladagur", (f"{year}-12-25" for year in self.full_range))

    def test_christmas_second_day(self):
        self.assertHolidayName("Annar jóladagur", (f"{year}-12-26" for year in self.full_range))

    def test_new_years_eve(self):
        self.assertHolidayName("Nýggjársaftan", (f"{year}-12-31" for year in self.full_range))

    def test_national_flag_day(self):
        name = "Flaggdagur"
        self.assertNoHolidayName(name)
        self.assertHalfDayHolidayName(name, (f"{year}-04-25" for year in self.full_range))

    def test_constitution_day(self):
        name = "Grundlógardagur"
        self.assertNoHolidayName(name)
        self.assertHalfDayHolidayName(name, (f"{year}-06-05" for year in self.full_range))

    def test_saint_olafs_eve(self):
        name = "Ólavsøkuaftan"
        self.assertNoHolidayName(name)
        self.assertHalfDayHolidayName(name, (f"{year}-07-28" for year in self.full_range))

    def test_l10n_default(self):
        self.assertLocalizedHolidays(
            ("2025-01-01", "Nýggjársdagur"),
            ("2025-04-17", "Skírhósdagur"),
            ("2025-04-18", "Langifríggjadagur"),
            ("2025-04-20", "Páskadagur"),
            ("2025-04-21", "Annar páskadagur"),
            ("2025-04-25", "Flaggdagur"),
            ("2025-05-16", "Dýri biðidagur"),
            ("2025-05-29", "Kristi himmalsferðardagur"),
            ("2025-06-05", "Grundlógardagur"),
            ("2025-06-08", "Hvítusunnudagur"),
            ("2025-06-09", "Annar hvítusunnudagur"),
            ("2025-07-28", "Ólavsøkuaftan"),
            ("2025-07-29", "Ólavsøkudagur"),
            ("2025-12-24", "Jólaaftan"),
            ("2025-12-25", "Jóladagur"),
            ("2025-12-26", "Annar jóladagur"),
            ("2025-12-31", "Nýggjársaftan"),
        )

    def test_l10n_da(self):
        self.assertLocalizedHolidays(
            "da",
            ("2025-01-01", "Nytårsdag"),
            ("2025-04-17", "Skærtorsdag"),
            ("2025-04-18", "Langfredag"),
            ("2025-04-20", "Påskedag"),
            ("2025-04-21", "Anden påskedag"),
            ("2025-04-25", "Flagdag"),
            ("2025-05-16", "Store bededag"),
            ("2025-05-29", "Kristi himmelfartsdag"),
            ("2025-06-05", "Grundlovsdag"),
            ("2025-06-08", "Pinsedag"),
            ("2025-06-09", "Anden pinsedag"),
            ("2025-07-28", "Olavsaften"),
            ("2025-07-29", "Olavsdag"),
            ("2025-12-24", "Juleaftensdag"),
            ("2025-12-25", "Juledag"),
            ("2025-12-26", "Anden juledag"),
            ("2025-12-31", "Nytårsaften"),
        )

    def test_l10n_en_us(self):
        self.assertLocalizedHolidays(
            "en_US",
            ("2025-01-01", "New Year's Day"),
            ("2025-04-17", "Maundy Thursday"),
            ("2025-04-18", "Good Friday"),
            ("2025-04-20", "Easter Sunday"),
            ("2025-04-21", "Easter Monday"),
            ("2025-04-25", "National Flag Day"),
            ("2025-05-16", "Great Prayer Day"),
            ("2025-05-29", "Ascension Day"),
            ("2025-06-05", "Constitution Day"),
            ("2025-06-08", "Whit Sunday"),
            ("2025-06-09", "Whit Monday"),
            ("2025-07-28", "Saint Olaf's Eve"),
            ("2025-07-29", "Saint Olaf's Day"),
            ("2025-12-24", "Christmas Eve"),
            ("2025-12-25", "Christmas Day"),
            ("2025-12-26", "Second Day of Christmas"),
            ("2025-12-31", "New Year's Eve"),
        )

    def test_l10n_is(self):
        self.assertLocalizedHolidays(
            "is",
            ("2025-01-01", "Nýársdagur"),
            ("2025-04-17", "Skírdagur"),
            ("2025-04-18", "Föstudagurinn langi"),
            ("2025-04-20", "Páskadagur"),
            ("2025-04-21", "Annar í páskum"),
            ("2025-04-25", "Fánadagur"),
            ("2025-05-16", "Kóngsbænadagur"),
            ("2025-05-29", "Uppstigningardagur"),
            ("2025-06-05", "Stjórnarskrárdagur"),
            ("2025-06-08", "Hvítasunnudagur"),
            ("2025-06-09", "Annar í hvítasunnu"),
            ("2025-07-28", "Ólafsnótt"),
            ("2025-07-29", "Ólafsdagur"),
            ("2025-12-24", "Aðfangadagur"),
            ("2025-12-25", "Jóladagur"),
            ("2025-12-26", "Annar í jólum"),
            ("2025-12-31", "Gamlársdagur"),
        )

    def test_l10n_no(self):
        self.assertLocalizedHolidays(
            "no",
            ("2025-01-01", "Nyttårsdag"),
            ("2025-04-17", "Skjærtorsdag"),
            ("2025-04-18", "Langfredag"),
            ("2025-04-20", "Første påskedag"),
            ("2025-04-21", "Andre påskedag"),
            ("2025-04-25", "Flaggdag"),
            ("2025-05-16", "Store bededag"),
            ("2025-05-29", "Kristi himmelfartsdag"),
            ("2025-06-05", "Grunnlovsdag"),
            ("2025-06-08", "Første pinsedag"),
            ("2025-06-09", "Andre pinsedag"),
            ("2025-07-28", "Olavsaften"),
            ("2025-07-29", "Olavsdag"),
            ("2025-12-24", "Julaften"),
            ("2025-12-25", "Første juledag"),
            ("2025-12-26", "Andre juledag"),
            ("2025-12-31", "Nyttårsaften"),
        )

    def test_l10n_sv(self):
        self.assertLocalizedHolidays(
            "sv",
            ("2025-01-01", "Nyårsdagen"),
            ("2025-04-17", "Skärtorsdagen"),
            ("2025-04-18", "Långfredagen"),
            ("2025-04-20", "Påskdagen"),
            ("2025-04-21", "Annandag påsk"),
            ("2025-04-25", "Flaggdagen"),
            ("2025-05-16", "Stora bönedagen"),
            ("2025-05-29", "Kristi himmelsfärdsdag"),
            ("2025-06-05", "Grundlagsdag"),
            ("2025-06-08", "Pingstdagen"),
            ("2025-06-09", "Annandag pingst"),
            ("2025-07-28", "Olafsafton"),
            ("2025-07-29", "Olafsdagen"),
            ("2025-12-24", "Julafton"),
            ("2025-12-25", "Juldagen"),
            ("2025-12-26", "Annandag jul"),
            ("2025-12-31", "Nyårsafton"),
        )



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

import pickle
import unittest
from datetime import date, datetime
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, FEB, OCT, DEC, MON, TUE, SAT, SUN
from holidays.constants import HOLIDAY_NAME_DELIMITER, OPTIONAL, PUBLIC, SCHOOL
from holidays.countries import UA, US
from holidays.groups.christian import ChristianHolidays
from holidays.groups.custom import StaticHolidays
from holidays.holiday_base import HolidayBase


class EntityStubStaticHolidays:
    special_public_holidays = {
        1111: (JAN, 1, "Test holiday"),
        2222: (FEB, 2, "Test holiday"),
        3333: ((FEB, 2, "Test holiday")),
        4444: (),
        1991: (
            (JAN, 7, JAN, 12),
            (JAN, 8, JAN, 13, 1991),
        ),
        2024: (
            (FEB, 19, FEB, 24),
            (OCT, 7, OCT, 12),
        ),
    }
    substituted_date_format = "%d/%m/%Y"
    substituted_label = "From %s"


class EntityStub(HolidayBase):
    def _add_observed(self, dt: date, *, before: bool = True, after: bool = True) -> None:
        if not self.observed:
            return None

        if self._is_saturday(dt) and before:
            self._add_holiday(f"{self[dt]} (observed)", dt + td(days=-1))
        elif self._is_sunday(dt) and after:
            self._add_holiday(f"{self[dt]} (observed)", dt + td(days=+1))

    def _populate(self, year: int) -> None:
        super()._populate(year)

        name = "New Year's Day"
        self._add_observed(self._add_holiday_jan_1(name), before=False)
        if self.observed and self._is_friday(DEC, 31):
            self._add_holiday_dec_31(f"{name} (observed)")

        self._add_observed(self._add_holiday_jun_19("Juneteenth National Independence Day"))
        self._add_observed(self._add_holiday_jul_4("Independence Day"))
        self._add_holiday_4th_thu_of_nov("Thanksgiving")
        self._add_observed(self._add_holiday_dec_25("Christmas Day"))


class CountryStub1(EntityStub, StaticHolidays):
    country = "CS1"
    subdivisions = ("Subdiv 1", "Subdiv 2", "3")
    subdivisions_aliases = {
        "S1": "Subdiv 1",
        "S2": "Subdiv 2",
        "S3": "3",
        "S_1": "Subdiv 1",
        "S_2": "Subdiv 2",
        "S_3": "3",
    }
    supported_categories = (PUBLIC, SCHOOL)
    start_year = 1000
    end_year = 3400

    def __init__(self, *args, **kwargs) -> None:
        StaticHolidays.__init__(self, cls=EntityStubStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_subdiv_subdiv_1_public_holidays(self):
        self._add_holiday_aug_10("Subdiv 1 Custom Holiday")

    def _populate_subdiv_subdiv_2_public_holidays(self):
        self._add_holiday_aug_10("Subdiv 2 Custom Holiday")


class CountryStub2(EntityStub):
    country = "CS2"
    subdivisions = ("Subdiv-3", "Subdiv-4")

    def _populate(self, year: int) -> None:
        super()._populate(year)
        self._add_holiday_mar_1("Custom March 1st Holiday")


class CountryStub3(HolidayBase):
    country = "CS3"

    def _populate(self, year: int) -> None:
        super()._populate(year)
        self._add_holiday_may_1("Custom May 1st Holiday")
        self._add_holiday_may_2("Custom May 2nd Holiday")


class CountryStub4(HolidayBase):
    country = "CS4"
    supported_categories = ("CUSTOM_1", "CUSTOM_2")


class CountryStub5(HolidayBase, ChristianHolidays):
    country = "CS5"

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        super().__init__(*args, **kwargs)


class CountryStub6(EntityStub, StaticHolidays):
    country = "CS6"

    def __init__(self, *args, **kwargs) -> None:
        StaticHolidays.__init__(self, cls=EntityStubStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self) -> None:
        self._add_holiday_may_1("Labor Day")
        self._add_holiday_may_2("Labor Day Two")
        self._add_holiday_oct_12("Columbus Day")


class MarketStub1(EntityStub):
    market = "MS1"


class MarketStub2(EntityStub):
    market = "MS2"


class TestArgs(unittest.TestCase):
    def test_categories(self):
        self.assertEqual({PUBLIC}, CountryStub1(categories=PUBLIC).categories)
        self.assertEqual({PUBLIC, SCHOOL}, CountryStub1(categories=(PUBLIC, SCHOOL)).categories)

        self.assertRaises(ValueError, lambda: CountryStub1(categories=OPTIONAL))
        self.assertRaises(ValueError, lambda: CountryStub1(categories="UNSUPPORTED"))
        self.assertRaises(ValueError, lambda: CountryStub1(categories=("HOME", "UNSUPPORTED")))
        self.assertRaises(ValueError, lambda: CountryStub4())

    def test_country(self):
        self.assertEqual(CountryStub1().country, "CS1")

    def test_expand(self):
        hb = CountryStub1(years=(2013, 2015), expand=False)
        self.assertFalse(hb.expand)
        self.assertSetEqual(hb.years, {2013, 2015})

        self.assertNotIn("2014-01-01", hb)
        self.assertSetEqual(hb.years, {2013, 2015})

    def test_observed(self):
        hb = CountryStub1(observed=False)
        self.assertFalse(hb.observed)
        self.assertIn("2000-01-01", hb)
        self.assertNotIn("1999-12-31", hb)
        self.assertIn("2012-01-01", hb)
        self.assertNotIn("2012-01-02", hb)

        hb.observed = True
        self.assertIn("2000-01-01", hb)
        self.assertIn("1999-12-31", hb)
        self.assertIn("2012-01-01", hb)
        self.assertIn("2012-01-02", hb)

        hb.observed = False
        self.assertIn("2000-01-01", hb)
        self.assertNotIn("1999-12-31", hb)
        self.assertIn("2012-01-01", hb)
        self.assertNotIn("2012-01-02", hb)

    def test_subdivision(self):
        self.assertEqual(CountryStub1(subdiv="Subdiv 1").subdiv, "Subdiv 1")
        self.assertEqual(CountryStub1(subdiv=3).subdiv, "3")

    def test_subdivisions_aliases(self):
        subdivisions_aliases = {
            "subdiv_1": ("S1", "S_1"),
            "subdiv_2": ("S2", "S_2"),
            "3": ("S3", "S_3"),
        }
        for subdiv, aliases in subdivisions_aliases.items():
            for alias in aliases:
                self.assertEqual(subdiv, CountryStub1(subdiv=alias)._normalized_subdiv)

        self.assertEqual(
            CountryStub1.get_subdivision_aliases(),
            {
                "Subdiv 1": ["S1", "S_1"],
                "Subdiv 2": ["S2", "S_2"],
                "3": ["S3", "S_3"],
            },
        )

    def test_years(self):
        hb = HolidayBase()
        self.assertSetEqual(hb.years, set())

        self.assertNotIn("2014-01-02", hb)
        self.assertSetEqual(hb.years, {2014})

        self.assertNotIn("2013-01-02", hb)
        self.assertNotIn("2014-01-02", hb)
        self.assertNotIn("2015-01-02", hb)
        self.assertSetEqual(hb.years, {2013, 2014, 2015})

        self.assertSetEqual(
            HolidayBase(years=range(2010, 2016)).years, {2010, 2011, 2012, 2013, 2014, 2015}
        )
        self.assertSetEqual(HolidayBase(years=(2013, 2015, 2015)).years, {2013, 2015})
        self.assertSetEqual(HolidayBase(years=(2013.0, 2015.0, 2015.0)).years, {2013, 2015})
        self.assertSetEqual(HolidayBase(years=2015).years, {2015})
        self.assertSetEqual(HolidayBase(years=2015.0).years, {2015})


class TestCategories(unittest.TestCase):
    class CustomCategoryClass(HolidayBase):
        country = "CCC"
        default_category = "CC"
        subdivisions = ("SD_1", "SD_2")
        supported_categories = ("CC", "CC_1", "CC_2")

        def _populate_cc_holidays(self):
            self._add_holiday_may_15("CC Holiday")

        def _populate_cc_1_holidays(self):
            self._add_holiday_may_16("CC1 Holiday")

        def _populate_cc_2_holidays(self):
            self._add_holiday_jun_15("CC2 Holiday")

        def _populate_subdiv_sd_1_cc_holidays(self):
            self._add_holiday_jul_1("SD_1 CC Holiday")

        def _populate_subdiv_sd_1_cc_1_holidays(self):
            self._add_holiday_jul_1("SD_1 CC_1 Holiday")

        def _populate_subdiv_sd_2_cc_holidays(self):
            self._add_holiday_aug_1("SD_2 CC Holiday")

    def test_default_category(self):
        ccc = TestCategories.CustomCategoryClass(years=2024)
        self.assertEqual(ccc.categories, {TestCategories.CustomCategoryClass.default_category})
        for name in ("CC Holiday",):
            self.assertTrue(ccc.get_named(name, lookup="exact"))
        for name in ("CC1 Holiday", "CC2 Holiday", "SD_1 CC_1 Holiday", "SD_2 CC Holiday"):
            self.assertFalse(ccc.get_named(name, lookup="exact"))

        # Default category with subdiv.
        ccc_sd2 = TestCategories.CustomCategoryClass(years=2024, subdiv="SD_2")
        self.assertEqual(ccc_sd2.categories, {TestCategories.CustomCategoryClass.default_category})
        for name in ("CC Holiday", "SD_2 CC Holiday"):
            self.assertTrue(ccc_sd2.get_named(name, lookup="exact"), name)
        for name in ("CC1 Holiday", "CC2 Holiday", "SD_1 CC_1 Holiday"):
            self.assertFalse(ccc_sd2.get_named(name, lookup="exact"), name)

    def test_no_default_category(self):
        TestCategories.CustomCategoryClass.default_category = None
        self.assertRaises(ValueError, lambda: TestCategories.CustomCategoryClass(years=2024))
        self.assertRaises(
            ValueError, lambda: TestCategories.CustomCategoryClass(years=2024, subdiv="SD_1")
        )

        # Explicitly set category.
        ccc = TestCategories.CustomCategoryClass(years=2024, categories="CC")
        self.assertEqual(ccc.categories, {"CC"})
        self.assertIn("2024-05-15", ccc)

    def test_non_default_categories(self):
        for categories in (
            {"CC_1": ("2024-05-16",)},
            {"CC_2": ("2024-06-15",)},
        ):
            for category in categories:
                ccc = TestCategories.CustomCategoryClass(years=2024, categories=category)
                self.assertEqual(ccc.categories, set(categories), category)
                for dt in categories[category]:
                    self.assertIn(dt, ccc)


class TestDeprecationWarnings(unittest.TestCase):
    def test_prov_deprecation(self):
        with self.assertWarns(Warning):
            CountryStub1(prov="Subdiv 1")

    def test_state_deprecation(self):
        with self.assertWarns(Warning):
            CountryStub1(state="Subdiv 1")


class TestEqualityInequality(unittest.TestCase):
    def test_eq(self):
        hb_1 = CountryStub1()
        hb_2 = CountryStub2()
        hb_3 = CountryStub1(subdiv="Subdiv 1")

        self.assertEqual(hb_1, hb_1)
        self.assertEqual(hb_2, hb_2)
        self.assertEqual(hb_3, hb_3)

        hb_4 = CountryStub1(years=2014)
        hb_5 = CountryStub1(years=2014)
        self.assertEqual(hb_4, hb_5)
        self.assertEqual(hb_5, hb_4)

        hb_6 = CountryStub2(years=(2014, 2015))
        hb_7 = CountryStub2(years=(2014, 2015))
        self.assertEqual(hb_6, hb_7)
        self.assertEqual(hb_7, hb_6)

        hb_8 = CountryStub3(language="fr")
        hb_9 = CountryStub3(language="fr")
        self.assertEqual(hb_8, hb_9)
        self.assertEqual(hb_9, hb_8)

        # Use assertFalse instead of assertNotEqual as we want to check "==" explicitly.
        self.assertFalse(hb_1 == {})
        self.assertFalse(hb_2 == {})
        self.assertFalse(hb_3 == {})
        self.assertFalse(hb_1 == hb_2)
        self.assertFalse(hb_2 == hb_1)
        self.assertFalse(hb_1 == hb_3)
        self.assertFalse(hb_3 == hb_1)
        self.assertFalse(hb_2 == hb_3)
        self.assertFalse(hb_3 == hb_2)

    def test_ne(self):
        hb_1 = CountryStub1()
        hb_2 = CountryStub2()
        hb_3 = CountryStub1(subdiv="Subdiv 1")

        self.assertNotEqual(hb_1, {})
        self.assertNotEqual(hb_2, {})
        self.assertNotEqual(hb_3, {})
        self.assertNotEqual(hb_1, hb_2)
        self.assertNotEqual(hb_2, hb_1)
        self.assertNotEqual(hb_1, hb_3)
        self.assertNotEqual(hb_3, hb_1)
        self.assertNotEqual(hb_2, hb_3)
        self.assertNotEqual(hb_3, hb_2)

        hb_4 = CountryStub1(years=2014)
        hb_5 = CountryStub2(years=2015)
        self.assertNotEqual(hb_1, hb_4)
        self.assertNotEqual(hb_4, hb_1)
        self.assertNotEqual(hb_2, hb_5)
        self.assertNotEqual(hb_5, hb_2)
        self.assertNotEqual(hb_4, hb_5)
        self.assertNotEqual(hb_5, hb_4)

        hb_6 = CountryStub1(years=2014, subdiv="Subdiv 1")
        hb_7 = CountryStub1(years=2014, subdiv="Subdiv 2")
        self.assertNotEqual(hb_6, hb_7)
        self.assertNotEqual(hb_7, hb_6)

        hb_8 = CountryStub1(years=2014)
        hb_9 = CountryStub1(years=2014, language="fr")
        self.assertNotEqual(hb_8, hb_9)
        self.assertNotEqual(hb_9, hb_8)

        # Use assertFalse instead of assertEqual in order to check "!=" explicitly.
        self.assertFalse(hb_1 != hb_1)
        self.assertFalse(hb_2 != hb_2)
        self.assertFalse(hb_3 != hb_3)


class TestGetList(unittest.TestCase):
    def test_get_list_multiple_countries(self):
        hb_country_1 = CountryStub1(years=2021)
        hb_country_2 = CountryStub1(years=2021)
        hb_country_1._add_holiday_dec_20("Custom Holiday 1")
        hb_country_2._add_holiday_dec_20("Custom Holiday 2")

        self.assertEqual(hb_country_1["2021-12-20"], "Custom Holiday 1")
        self.assertEqual(hb_country_2["2021-12-20"], "Custom Holiday 2")
        self.assertListEqual(hb_country_1.get_list("2021-12-20"), ["Custom Holiday 1"])
        self.assertListEqual(hb_country_2.get_list("2021-12-20"), ["Custom Holiday 2"])

        hb_combined = hb_country_1 + hb_country_2
        self.assertEqual(hb_combined["2021-12-20"], "Custom Holiday 1; Custom Holiday 2")
        self.assertListEqual(
            hb_combined.get_list("2021-12-20"), ["Custom Holiday 1", "Custom Holiday 2"]
        )

    def test_get_list_multiple_subdivisions(self):
        hb_subdiv_1 = CountryStub1(subdiv="Subdiv 1", years=2021)
        hb_subdiv_2 = CountryStub1(subdiv="Subdiv 2", years=2021)
        self.assertEqual(hb_subdiv_1["2021-08-10"], "Subdiv 1 Custom Holiday")
        self.assertEqual(hb_subdiv_2["2021-08-10"], "Subdiv 2 Custom Holiday")
        self.assertListEqual(hb_subdiv_1.get_list("2021-08-10"), ["Subdiv 1 Custom Holiday"])
        self.assertListEqual(hb_subdiv_2.get_list("2021-08-10"), ["Subdiv 2 Custom Holiday"])

        hb_combined = hb_subdiv_1 + hb_subdiv_2
        self.assertEqual(
            hb_combined["2021-08-10"], "Subdiv 1 Custom Holiday; Subdiv 2 Custom Holiday"
        )
        self.assertListEqual(
            hb_combined.get_list("2021-08-10"),
            ["Subdiv 1 Custom Holiday", "Subdiv 2 Custom Holiday"],
        )


class TestGetNamed(unittest.TestCase):
    def test_contains(self):
        hb = CountryStub1(years=2022)
        for name in ("New", "Year"):
            self.assertListEqual(hb.get_named(name, lookup="contains"), [date(2022, 1, 1)])
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="contains"),
            [date(2022, 6, 19), date(2022, 6, 20), date(2022, 7, 4)],
        )
        self.assertListEqual(hb.get_named("independence day", lookup="contains"), [])
        self.assertListEqual(hb.get_named("Thanksgiving", lookup="contains"), [date(2022, 11, 24)])
        self.assertListEqual(hb.get_named("Thanksgivi", lookup="contains"), [date(2022, 11, 24)])
        self.assertListEqual(hb.get_named("thanks", lookup="contains"), [])
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="contains"),
            [date(2022, 6, 19), date(2022, 7, 4)],
        )
        self.assertListEqual(hb.get_named("independence day", lookup="contains"), [])
        self.assertSetEqual(hb.years, {2022})

    def test_exact(self):
        hb = CountryStub1(years=2022)
        for name in ("New", "Day"):
            self.assertListEqual(hb.get_named(name, lookup="exact"), [])
        self.assertListEqual(hb.get_named("Independence Day", lookup="exact"), [date(2022, 7, 4)])
        self.assertListEqual(hb.get_named("Thanksgiving", lookup="exact"), [date(2022, 11, 24)])
        self.assertListEqual(hb.get_named("thanksgiving", lookup="exact"), [])
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(hb.get_named("Independence Day", lookup="exact"), [date(2022, 7, 4)])
        self.assertSetEqual(hb.years, {2022})

    def test_icontains(self):
        hb = CountryStub1(years=2022)
        for name in ("Thanksgiving", "thanksgiving", "Thanksgivi"):
            self.assertListEqual(hb.get_named(name, lookup="icontains"), [date(2022, 11, 24)])
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="icontains"),
            [date(2022, 6, 19), date(2022, 6, 20), date(2022, 7, 4)],
        )
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="icontains"),
            [date(2022, 6, 19), date(2022, 7, 4)],
        )
        self.assertSetEqual(hb.years, {2022})

    def test_iexact(self):
        hb = CountryStub1(years=2022)
        for name in ("new year's day", "New Year's Day"):
            self.assertListEqual(hb.get_named(name, lookup="iexact"), [date(2022, 1, 1)])
        for name in ("New Year Day", "New Year", "year", "NEW Year"):
            self.assertListEqual(hb.get_named(name, lookup="iexact"), [])
        self.assertListEqual(hb.get_named("independence day", lookup="iexact"), [date(2022, 7, 4)])
        self.assertListEqual(hb.get_named("thanksgiving", lookup="iexact"), [date(2022, 11, 24)])
        self.assertListEqual(hb.get_named("Thanksgivin", lookup="iexact"), [])
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(hb.get_named("independence day", lookup="iexact"), [date(2022, 7, 4)])
        self.assertSetEqual(hb.years, {2022})

    def test_invalid(self):
        hb = CountryStub1(years=2022)
        self.assertRaises(
            AttributeError, lambda: hb.get_named("Independence Day", lookup="invalid")
        )

    def test_istartswith(self):
        hb = CountryStub1(years=2022)
        for name in ("new year's", "New Year's", "New Year's day"):
            self.assertListEqual(hb.get_named(name, lookup="istartswith"), [date(2022, 1, 1)])
        for name in ("New Year Day", "New Year holiday", "New Year's Day Holiday", "year"):
            self.assertListEqual(hb.get_named(name, lookup="istartswith"), [])
        self.assertListEqual(
            hb.get_named("independence day", lookup="istartswith"), [date(2022, 7, 4)]
        )
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(
            hb.get_named("independence day", lookup="istartswith"), [date(2022, 7, 4)]
        )
        self.assertSetEqual(hb.years, {2022})

    def test_startswith(self):
        hb = CountryStub1(years=2022)
        for name in ("New Year's", "New Year"):
            self.assertListEqual(hb.get_named(name, lookup="startswith"), [date(2022, 1, 1)])
        for name in ("New Year Day", "New Year Holiday", "New Year's Day Holiday", "year"):
            self.assertListEqual(hb.get_named(name, lookup="startswith"), [])
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="startswith"), [date(2022, 7, 4)]
        )
        self.assertListEqual(
            hb.get_named("Christmas", lookup="startswith"),
            [date(2022, 12, 25), date(2022, 12, 26)],
        )
        self.assertSetEqual(hb.years, {2022})

        hb = CountryStub1(observed=False, years=2022)
        self.assertListEqual(
            hb.get_named("Independence Day", lookup="startswith"), [date(2022, 7, 4)]
        )
        self.assertListEqual(hb.get_named("Christmas", lookup="startswith"), [date(2022, 12, 25)])
        self.assertSetEqual(hb.years, {2022})


class TestHelperMethods(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_add_holiday(self):
        self.hb._populate(2023)
        self.hb._add_holiday("Test 1", date(2023, JAN, 5))
        self.hb._add_holiday("Test 2", (JAN, 6))
        self.hb._add_holiday_jan_7("Test 3")

        self.assertIn("2023-01-05", self.hb)
        self.assertIn("2023-01-06", self.hb)
        self.assertIn("2023-01-07", self.hb)

        for args in (
            (date(2020, JAN, 5),),
            (JAN, 5, "Test 1", True),
            ("Test", "Test"),
        ):
            self.assertRaises(TypeError, lambda: self.hb._add_holiday(*args))

    def test_add_multiday_holiday(self):
        self.hb._populate(2024)

        name = "Multi-day Holiday"
        self.hb._add_multiday_holiday(self.hb._add_holiday(name, (JAN, 1)), 5)
        self.assertIn("2024-01-01", self.hb)
        self.assertIn("2024-01-02", self.hb)
        self.assertIn("2024-01-03", self.hb)
        self.assertIn("2024-01-04", self.hb)
        self.assertIn("2024-01-05", self.hb)
        self.assertIn("2024-01-06", self.hb)
        self.assertEqual(len(self.hb.get_named(name)), 6)

        name_2 = "Multi-day Holiday 2"
        self.hb._add_multiday_holiday(self.hb._add_holiday(name, (DEC, 25)), 6, name=name_2)
        self.assertIn("2024-12-25", self.hb)
        self.assertIn("2024-12-26", self.hb)
        self.assertIn("2024-12-27", self.hb)
        self.assertIn("2024-12-28", self.hb)
        self.assertIn("2024-12-29", self.hb)
        self.assertIn("2024-12-30", self.hb)
        self.assertIn("2024-12-31", self.hb)
        # DEC, 25 itself is name instead of name_2.
        self.assertEqual(len(self.hb.get_named(name_2)), 6)

    def test_add_multiday_holiday_invalid_args(self):
        self.hb._populate(2024)
        with self.assertRaises(ValueError):
            self.hb._add_multiday_holiday(date(2024, DEC, 1), 3)

    def test_is_leap_year(self):
        self.hb._populate(1999)
        self.assertFalse(self.hb._is_leap_year())

        self.hb._populate(2000)
        self.assertTrue(self.hb._is_leap_year())

        self.hb._populate(2004)
        self.assertTrue(self.hb._is_leap_year())

        self.hb._populate(2200)
        self.assertFalse(self.hb._is_leap_year())

    def test_is_weekend(self):
        self.hb._populate(2022)
        dts = (date(2022, 10, 3), date(2022, 10, 4), (OCT, 3), (OCT, 4))

        self.hb.weekend = {MON, TUE}
        for dt in dts:
            self.assertTrue(self.hb._is_weekend(dt))
            self.assertFalse(self.hb._is_weekday(dt))

        self.hb.weekend = {}
        for dt in dts:
            self.assertFalse(self.hb._is_weekend(dt))
            self.assertTrue(self.hb._is_weekday(dt))

        self.hb.weekend = {SAT, SUN}
        for dt in (date(2022, 10, 1), date(2022, 10, 2)):
            self.assertTrue(self.hb._is_weekend(dt))
            self.assertFalse(self.hb._is_weekday(dt))
        for dt in ((OCT, 1), (OCT, 2)):
            self.assertTrue(self.hb._is_weekend(dt))
            self.assertTrue(self.hb._is_weekend(*dt))
            self.assertFalse(self.hb._is_weekday(dt))
            self.assertFalse(self.hb._is_weekday(*dt))

        for dt in (date(2022, 10, 3), date(2022, 10, 4)):
            self.assertFalse(self.hb._is_weekend(dt))
            self.assertTrue(self.hb._is_weekday(dt))
        for dt in ((OCT, 3), (OCT, 4)):
            self.assertFalse(self.hb._is_weekend(dt))
            self.assertFalse(self.hb._is_weekend(*dt))
            self.assertTrue(self.hb._is_weekday(dt))
            self.assertTrue(self.hb._is_weekday(*dt))


class TestHolidaySum(unittest.TestCase):
    def setUp(self) -> None:
        self.hb_1 = CountryStub1(years=2014)
        self.hb_2 = CountryStub2(years=2015)
        self.hb_3 = CountryStub3()
        self.hb_combined = self.hb_1 + self.hb_2 + self.hb_3

    def assertAdded(self):  # noqa: N802
        self.assertNotIn("2014-03-01", self.hb_1)
        self.assertNotIn("2014-05-01", self.hb_1)
        self.assertNotIn("2014-05-02", self.hb_1)
        self.assertIn("2014-03-01", self.hb_2)
        self.assertIn("2014-05-01", self.hb_3)
        self.assertIn("2014-05-02", self.hb_3)

        self.assertIn("2014-07-04", self.hb_1)
        self.assertIn("2014-07-04", self.hb_1 + self.hb_2)
        self.assertIn("2014-07-04", self.hb_2 + self.hb_1)
        self.assertIn("2015-07-04", self.hb_1 + self.hb_2)
        self.assertIn("2015-07-04", self.hb_2 + self.hb_1)

        self.assertIn("2000-03-01", self.hb_combined)
        self.assertIn("2000-05-01", self.hb_combined)
        self.assertIn("2000-05-02", self.hb_combined)
        self.assertIn("2000-07-04", self.hb_combined)
        self.assertIn("2000-07-04", self.hb_combined)

    def test_add_country(self):
        self.assertAdded()

        self.hb_combined = CountryStub1(years=2014)
        self.hb_combined += CountryStub2(years=2015)
        self.hb_combined += CountryStub3()
        self.assertAdded()

        self.hb_combined = (
            self.hb_1
            + (self.hb_2 + self.hb_3)
            + self.hb_1
            + (self.hb_3 + self.hb_2 + CountryStub2(subdiv="Subdiv-4"))
        )
        self.assertAdded()

        self.hb_combined = CountryStub1(years=2014, subdiv="Subdiv 1")
        self.hb_combined += CountryStub1(years=2014, subdiv="Subdiv 2")
        self.assertEqual(
            self.hb_combined["2014-08-10"], "Subdiv 1 Custom Holiday; Subdiv 2 Custom Holiday"
        )

        self.assertRaises(TypeError, lambda: self.hb_1 + {})

    def test_add_financial(self):
        hb_1 = MarketStub1(years=2013)
        hb_2 = MarketStub2(years=(2014, 2015))
        hb_combined = hb_1 + hb_2

        self.assertIn("2013-01-01", hb_1)
        self.assertIn("2014-01-01", hb_2)
        self.assertIn("2015-01-01", hb_2)
        self.assertIn("2013-01-01", hb_combined)
        self.assertIn("2014-01-01", hb_combined)
        self.assertIn("2015-01-01", hb_combined)
        self.assertListEqual(hb_combined.market, ["MS1", "MS2"])

    def test_args(self):
        self.hb_1 = CountryStub1(years=2014, subdiv="Subdiv 1")
        self.hb_2 = CountryStub2(years=2015, subdiv="Subdiv-3")
        self.assertListEqual(self.hb_combined.country, ["CS1", "CS2", "CS3"])
        self.assertListEqual((self.hb_1 + self.hb_2).subdiv, ["Subdiv 1", "Subdiv-3"])
        self.assertListEqual((self.hb_2 + self.hb_1).subdiv, ["Subdiv-3", "Subdiv 1"])
        self.assertEqual((self.hb_1 + self.hb_3).subdiv, "Subdiv 1")

        self.assertTrue((self.hb_2 + self.hb_3).expand)
        self.assertSetEqual((self.hb_2 + self.hb_3).years, {2014, 2015})
        self.assertSetEqual((self.hb_3 + self.hb_2).years, {2014, 2015})

        self.hb_combined = sum(CountryStub1(subdiv=subdiv) for subdiv in CountryStub1.subdivisions)
        self.assertEqual(self.hb_combined.country, CountryStub1.country)
        self.assertEqual(self.hb_combined.subdiv, list(CountryStub1.subdivisions))


class TestInheritance(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_custom_holidays(self):
        class CustomHolidays(CountryStub1):
            def _populate(self, year):
                super()._populate(year)
                self._add_holiday_jul_13("Ninja Turtle's Day")
                self._add_holiday_dec_31("New Year's Eve")
                del self[date(year, JAN, 1)]

        hb = CustomHolidays(years=(2014, 2020))
        self.assertIn("2014-01-01", self.hb)
        self.assertNotIn("2014-01-01", hb)
        self.assertIn("2014-12-31", hb)
        self.assertNotIn("2014-12-31", self.hb)

        self.assertIn("2020-01-01", self.hb)
        self.assertNotIn("2020-01-01", hb)
        self.assertIn("2020-12-31", hb)
        self.assertNotIn("2020-12-31", self.hb)

        self.assertNotIn("2014-07-13", self.hb)
        self.assertIn("2014-07-13", hb)
        self.assertNotIn("2020-07-13", self.hb)
        self.assertIn("2020-07-13", hb)


class TestIsWeekend(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_is_weekend(self):
        self.hb._populate(2022)

        self.hb.weekend = {MON, TUE}
        for dt in (date(2022, 10, 3), date(2022, 10, 4), "2022-10-03", "2022-10-04"):
            self.assertTrue(self.hb.is_weekend(dt))

        self.hb.weekend = set()
        for dt in (date(2022, 10, 3), date(2022, 10, 4), "2022-10-03", "2022-10-04"):
            self.assertFalse(self.hb.is_weekend(dt))

        self.hb.weekend = {SAT, SUN}
        for dt in (date(2022, 10, 1), date(2022, 10, 2), "2022-10-01", "2022-10-02"):
            self.assertTrue(self.hb.is_weekend(dt))

        for dt in (date(2022, 10, 3), date(2022, 10, 4), "2022-10-03", "2022-10-04"):
            self.assertFalse(self.hb.is_weekend(dt))


class TestKeyTransforms(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_date(self):
        dt = date(2014, 1, 1)
        self.assertIn(dt, self.hb)
        self.assertEqual(self.hb[dt], "New Year's Day")

        dt = date(2014, 1, 3)
        self.hb[dt] = "Fake Holiday"
        self.assertIn(dt, self.hb)
        self.assertEqual(self.hb.pop(dt), "Fake Holiday")
        self.assertNotIn(dt, self.hb)

    def test_date_subclass(self):
        class CustomDateType(date):
            pass

        self.assertTrue(issubclass(CustomDateType, date))
        self.assertIn(CustomDateType(2014, 1, 1), self.hb)
        self.assertNotIn(CustomDateType(2014, 1, 3), self.hb)

    def test_datetime(self):
        self.assertIn(datetime(2014, 1, 1, 13, 45), self.hb)
        self.assertEqual(self.hb[datetime(2014, 1, 1, 13, 45)], "New Year's Day")

        self.hb[datetime(2014, 1, 3, 1, 1)] = "Fake Holiday"
        self.assertIn(datetime(2014, 1, 3, 2, 2), self.hb)
        self.assertEqual(self.hb.pop(datetime(2014, 1, 3, 4, 4)), "Fake Holiday")
        self.assertNotIn(datetime(2014, 1, 3, 2, 2), self.hb)

    def test_exception(self):
        self.assertRaises((TypeError, ValueError), lambda: "abc" in self.hb)
        self.assertRaises((TypeError, ValueError), lambda: self.hb.get("abc123"))
        self.assertRaises(TypeError, lambda: self.hb.get({"123"}))
        self.assertRaises((TypeError, ValueError), self.hb.__setitem__, "abc", "Test")
        self.assertRaises((TypeError, ValueError), lambda: {} in self.hb)

    def test_string(self):
        self.assertIn("2014-01-01", self.hb)
        self.assertEqual(self.hb["2014-01-01"], "New Year's Day")
        self.assertIn("01/01/2014", self.hb)
        self.assertEqual(self.hb["01/01/2014"], "New Year's Day")

        self.hb["01/03/2014"] = "Fake Holiday"
        self.assertIn("01/03/2014", self.hb)
        self.assertNotIn("03/01/2014", self.hb)
        self.assertIn("2014-01-03", self.hb)
        self.assertNotIn("2014-03-01", self.hb)
        self.assertEqual(self.hb.pop("01/03/2014"), "Fake Holiday")

    def test_timestamp(self):
        self.assertIn(1388552400, self.hb)
        self.assertEqual(self.hb[1388552400], "New Year's Day")
        self.assertIn(1388552400.01, self.hb)
        self.assertEqual(self.hb[1388552400.01], "New Year's Day")

        self.hb[1388725200] = "Fake Holiday"
        self.assertIn(1388725201, self.hb)
        self.assertEqual(self.hb.pop(1388725202), "Fake Holiday")
        self.assertNotIn(1388725201, self.hb)


class TestPop(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_exception(self):
        self.assertRaises(KeyError, lambda: self.hb.pop("2014-01-02"))

        self.hb.pop("2014-01-01")
        self.assertRaises(KeyError, lambda: self.hb.pop("2014-01-01"))

    def test_success(self):
        self.assertFalse(self.hb.pop("2014-01-02", False))
        self.assertTrue(self.hb.pop("2014-01-02", True))
        self.assertIn("2014-01-01", self.hb)
        self.assertEqual(self.hb.pop("2014-01-01"), "New Year's Day")
        self.assertNotIn("2014-01-01", self.hb)


class TestPopNamed(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_exception(self):
        self.assertRaises(KeyError, lambda: self.hb.pop_named("New Year's Dayz"))

        self.assertIn("2014-01-01", self.hb)
        self.hb.pop_named("New Year's Day")
        self.assertRaises(KeyError, lambda: self.hb.pop_named("New Year's Day"))

    def test_single(self):
        self.assertIn("2014-01-01", self.hb)
        for dt in self.hb.pop_named("New Year's Day"):
            self.assertNotIn(dt, self.hb)

    def test_multiple(self):
        dt = date(2022, 2, 22)
        holiday_name_1 = "Holiday Name 1"
        holiday_name_2 = "Holiday Name 2"
        holiday_name_3 = "Holiday Name 3"
        combined_name = HOLIDAY_NAME_DELIMITER.join(
            (holiday_name_1, holiday_name_2, holiday_name_3)
        )
        self.hb[dt] = holiday_name_1
        self.hb[dt] = holiday_name_2
        self.hb[dt] = holiday_name_3
        self.assertEqual(self.hb[dt], combined_name)

        # Pop the entire date by multiple holidays exact name.
        self.hb.pop_named(combined_name)
        self.assertNotIn(dt, self.hb)

        # Pop only one holiday by a single name.
        self.hb[dt] = holiday_name_1
        self.hb[dt] = holiday_name_2
        self.hb[dt] = holiday_name_3

        # 3 holidays names.
        self.assertEqual(self.hb[dt], combined_name)

        self.hb.pop_named(holiday_name_1)
        # 2 holiday names.
        self.assertEqual(
            self.hb[dt], HOLIDAY_NAME_DELIMITER.join((holiday_name_2, holiday_name_3))
        )

        self.hb.pop_named(holiday_name_3)
        # 1 holiday name.
        self.assertEqual(self.hb[dt], holiday_name_2)

        self.hb.pop_named(holiday_name_2)
        # 0 holiday names.
        self.assertNotIn(dt, self.hb)

    def test_partial(self):
        self.assertIn("2014-01-01", self.hb)
        for dt in self.hb.pop_named("N"):
            self.assertNotIn(dt, self.hb)
        self.assertRaises(KeyError, lambda: self.hb.pop_named("New Year"))

    def test_contains(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("Day", lookup="contains")
        self.assertEqual(len(removed_dates), 6)
        self.assertNotIn("2022-01-01", self.hb)
        self.assertNotIn("2022-06-19", self.hb)
        self.assertNotIn("2022-06-20", self.hb)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertNotIn("2022-12-25", self.hb)
        self.assertNotIn("2022-12-26", self.hb)

    def test_icontains(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("day", lookup="icontains")
        self.assertEqual(len(removed_dates), 6)
        self.assertNotIn("2022-01-01", self.hb)
        self.assertNotIn("2022-06-19", self.hb)
        self.assertNotIn("2022-06-20", self.hb)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertNotIn("2022-12-25", self.hb)
        self.assertNotIn("2022-12-26", self.hb)

    def test_exact(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("Independence Day", lookup="exact")
        self.assertEqual(len(removed_dates), 1)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertIn("2022-06-19", self.hb)
        self.assertIn("2022-06-20", self.hb)

        hb = CountryStub1(years=2025)
        hb["2024-02-02"] = "Big Groundhog Day; Groundhog Day"
        removed_dates = hb.pop_named("Groundhog Day", lookup="exact")
        self.assertEqual(len(removed_dates), 1)
        self.assertEqual(hb["2024-02-02"], "Big Groundhog Day")

    def test_iexact(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("independence day", lookup="iexact")
        self.assertEqual(len(removed_dates), 1)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertIn("2022-06-19", self.hb)
        self.assertIn("2022-06-20", self.hb)

        hb = CountryStub1(years=2025)
        hb["2024-02-02"] = "Big Groundhog Day; Groundhog Day"
        removed_dates = hb.pop_named("Groundhog day", lookup="iexact")
        self.assertEqual(len(removed_dates), 1)
        self.assertEqual(hb["2024-02-02"], "Big Groundhog Day")

    def test_startswith(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("Independence", lookup="startswith")
        self.assertEqual(len(removed_dates), 1)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertIn("2022-06-19", self.hb)
        self.assertIn("2022-06-20", self.hb)

        removed_dates = self.hb.pop_named("Christmas", lookup="startswith")
        self.assertEqual(len(removed_dates), 2)
        self.assertNotIn("2022-12-25", self.hb)
        self.assertNotIn("2022-12-26", self.hb)

    def test_istartswith(self):
        self.assertIn("2022-01-01", self.hb)
        removed_dates = self.hb.pop_named("independence", lookup="istartswith")
        self.assertEqual(len(removed_dates), 1)
        self.assertNotIn("2022-07-04", self.hb)
        self.assertIn("2022-06-19", self.hb)
        self.assertIn("2022-06-20", self.hb)

        removed_dates = self.hb.pop_named("christmas", lookup="istartswith")
        self.assertEqual(len(removed_dates), 2)
        self.assertNotIn("2022-12-25", self.hb)
        self.assertNotIn("2022-12-26", self.hb)


class TestRepr(unittest.TestCase):
    def test_base(self):
        self.assertEqual(repr(HolidayBase()), "holidays.HolidayBase()")

    def test_country(self):
        self.assertEqual(repr(CountryStub1()), "holidays.country_holidays('CS1')")
        self.assertEqual(
            repr(CountryStub1(subdiv="Subdiv 1")),
            "holidays.country_holidays('CS1', subdiv='Subdiv 1')",
        )
        self.assertEqual(
            repr(CountryStub1(language="de")), "holidays.country_holidays('CS1', language='de')"
        )
        self.assertEqual(
            repr(CountryStub1(categories=SCHOOL)),
            "holidays.country_holidays('CS1', categories=['school'])",
        )
        self.assertEqual(
            repr(CountryStub1(categories=(PUBLIC, SCHOOL))),
            "holidays.country_holidays('CS1', categories=['public', 'school'])",
        )
        self.assertEqual(
            repr(CountryStub1(language="de", subdiv="Subdiv 1")),
            "holidays.country_holidays('CS1', language='de', subdiv='Subdiv 1')",
        )
        self.assertEqual(
            repr(CountryStub1(categories=(PUBLIC, SCHOOL), language="de", subdiv="Subdiv 1")),
            "holidays.country_holidays('CS1', categories=['public', 'school'], "
            "language='de', subdiv='Subdiv 1')",
        )

    def test_market(self):
        self.assertEqual(repr(MarketStub1()), "holidays.financial_holidays('MS1')")
        self.assertEqual(
            repr(MarketStub1(language="de")), "holidays.financial_holidays('MS1', language='de')"
        )


class TestSerialization(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_pickle(self):
        dt = "2020-01-01"
        self.assertIn(dt, self.hb)

        loaded_holidays = pickle.loads(pickle.dumps(self.hb))
        self.assertEqual(loaded_holidays, self.hb)
        self.assertIn(dt, self.hb)

    def test_pickle_localized_entity(self):
        for lang in ("uk", "en_US", None):
            ua = UA(language=lang)
            dt = "2021-01-01"
            self.assertIn(dt, self.hb)

            loaded_ua = pickle.loads(pickle.dumps(ua))
            self.assertEqual(loaded_ua, ua)
            self.assertIn(dt, loaded_ua)


class TestSpecialHolidays(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_populate_special_holidays(self):
        self.assertSetEqual(self.hb.years, set())

        self.hb._populate(1111)
        self.assertSetEqual(self.hb.years, {1111})
        self.assertIn("1111-01-01", self.hb)
        self.assertIn("2222-02-02", self.hb)
        self.assertIn("3333-02-02", self.hb)
        self.assertSetEqual(self.hb.years, {1111, 2222, 3333})


class TestStandardMethods(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub1()

    def test_append(self):
        self.hb.append([date(2015, 4, 1), "2015-04-03"])
        self.hb.append(date(2015, 4, 6))
        self.hb.append("2015-04-07")

        for dt in (
            "2015-01-01",
            "2015-04-01",
            "2015-04-03",
            "2015-04-06",
            "2015-04-07",
            "2015-12-25",
        ):
            self.assertIn(dt, self.hb)

        for dt in (
            "2015-04-04",
            "2015-04-05",
            "2015-04-02",
        ):
            self.assertNotIn(dt, self.hb)

    def test_bool(self):
        self.assertFalse(self.hb)
        self.assertEqual(bool(self.hb), False)
        self.assertEqual(len(self.hb), 0)

        self.assertIn("2014-01-01", self.hb)
        self.assertTrue(self.hb)
        self.assertEqual(bool(self.hb), True)
        self.assertNotEqual(len(self.hb), 0)

    def test_contains(self):
        self.assertIn("2014-01-01", self.hb)
        self.assertNotIn("2014-01-03", self.hb)

    def test_copy(self):
        hb = CountryStub1()
        self.assertEqual(hb, hb.copy())

        hb_fr = CountryStub1(language="fr")
        hb_xx = CountryStub1(language="xx")
        self.assertNotEqual(hb, hb_fr)
        self.assertNotEqual(hb.copy(), hb_fr.copy())

        self.assertNotEqual(hb, hb_xx)
        self.assertNotEqual(hb.copy(), hb_xx.copy())

    def test_get(self):
        self.assertEqual(self.hb.get("2014-01-01"), "New Year's Day")
        self.assertIsNone(self.hb.get("2014-01-03"))
        self.assertFalse(self.hb.get("2014-01-03", False))
        self.assertTrue(self.hb.get("2014-01-03", True))

    def test_getattr(self):
        self.hb._populate(2023)

        name = "Test"
        self.assertEqual(self.hb._add_holiday_3rd_mon_of_jun(name), date(2023, 6, 19))
        self.assertEqual(self.hb._add_holiday_4th_fri_of_aug(name), date(2023, 8, 25))
        self.assertEqual(self.hb._add_holiday_1st_wed_of_aug(name), date(2023, 8, 2))
        self.assertEqual(self.hb._add_holiday_1st_mon_before_may_24(name), date(2023, 5, 22))
        self.assertEqual(self.hb._add_holiday_1st_sun_from_aug_31(name), date(2023, 9, 3))

        h_chr = CountryStub5()
        h_chr._populate(2023)
        self.assertEqual(h_chr._add_holiday_8_days_past_easter(name), date(2023, 4, 17))
        self.assertEqual(h_chr._add_holiday_5_days_prior_easter(name), date(2023, 4, 4))

        self.assertRaises(ValueError, lambda: self.hb._add_holiday_5th_fri_of_aug(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_4th_nam_of_aug(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_nam_12(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_1st_fri_before_nam_29(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_1st_fri_random_jan_29(name))
        self.assertRaises(AttributeError, lambda: h_chr._add_holiday_18_days_past_eastr(name))

        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_fe_15(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_2_may_15(name))
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_3th_su_of_aug(name))
        self.assertRaises(
            AttributeError, lambda: self.hb._add_holiday_2_days_past_1st_sat_of_ap(name)
        )
        self.assertRaises(AttributeError, lambda: self.hb._add_holiday_1st_sat_from_fe_10(name))

    def test_getitem(self):
        self.assertEqual(self.hb["2014-01-01"], "New Year's Day")
        self.assertEqual(self.hb.get("2014-01-01"), "New Year's Day")
        self.assertIsNone(self.hb.get("2014-01-03"))
        self.assertRaises(KeyError, lambda: self.hb["2014-01-03"])

        self.assertListEqual(self.hb["2013-12-31":"2014-01-02"], [date(2014, 1, 1)])
        self.assertListEqual(
            self.hb["2013-12-24":"2014-01-02"], [date(2013, 12, 25), date(2014, 1, 1)]
        )
        self.assertListEqual(self.hb["2013-12-25":"2014-01-02":3], [date(2013, 12, 25)])
        self.assertListEqual(
            self.hb["2013-12-25":"2014-01-02":7], [date(2013, 12, 25), date(2014, 1, 1)]
        )
        self.assertListEqual(self.hb["2014-01-02":"2013-12-30"], [date(2014, 1, 1)])
        self.assertListEqual(self.hb["2014-01-02":"2013-12-25"], [date(2014, 1, 1)])
        self.assertListEqual(
            self.hb["2014-01-02":"2013-12-24"], [date(2014, 1, 1), date(2013, 12, 25)]
        )
        self.assertListEqual(self.hb["2014-01-01":"2013-12-24":3], [date(2014, 1, 1)])
        self.assertListEqual(
            self.hb["2014-01-01":"2013-12-24":7], [date(2014, 1, 1), date(2013, 12, 25)]
        )
        self.assertListEqual(self.hb["2013-12-31":"2014-01-02":-3], [])
        self.assertListEqual(self.hb["2014-01-01" : "2013-12-24" : td(days=3)], [date(2014, 1, 1)])
        self.assertListEqual(
            self.hb["2014-01-01" : "2013-12-24" : td(days=7)],
            [date(2014, 1, 1), date(2013, 12, 25)],
        )
        self.assertListEqual(self.hb["2013-12-31" : "2014-01-02" : td(days=3)], [])

        self.assertRaises(ValueError, lambda: self.hb["2014-01-01":])
        self.assertRaises(ValueError, lambda: self.hb[:"2014-01-01"])
        self.assertRaises(TypeError, lambda: self.hb["2014-01-01":"2014-01-02":""])
        self.assertRaises(ValueError, lambda: self.hb["2014-01-01":"2014-01-02":0])
        self.assertRaises(ValueError, lambda: self.hb["2014-01-01" : "2014-01-02" : td(hours=12)])

    def test_radd(self):
        self.assertRaises(TypeError, lambda: 1 + CountryStub1())

    def test_setitem(self):
        self.assertEqual(len(self.hb), 0)
        self.hb["2014-01-03"] = "Custom Holiday"
        self.assertGreater(len(self.hb), 0)
        self.assertIn("2014-01-03", self.hb)
        self.assertEqual(self.hb["2014-01-03"], "Custom Holiday")

        self.hb["2014-01-04"] = "Custom Holiday"
        self.hb["2014-01-04"] = "Custom Holiday; Another Custom Holiday"
        self.assertEqual(self.hb["2014-01-04"], "Another Custom Holiday; Custom Holiday")

    def test_update(self):
        self.hb.update(
            {
                "2015-01-10": "Custom Holiday",
                "2015-01-11": "Custom Holiday",
            }
        )
        self.assertIn("2015-01-10", self.hb)
        self.assertIn("2015-01-11", self.hb)
        self.assertEqual(self.hb["2015-01-10"], "Custom Holiday")
        self.assertEqual(self.hb["2015-01-11"], "Custom Holiday")

        self.hb.update(
            {
                date(2015, 1, 10): "New Holiday",
                date(2015, 1, 11): "New Holiday",
            }
        )
        self.assertIn(date(2015, 1, 10), self.hb)
        self.assertIn(date(2015, 1, 11), self.hb)
        self.assertEqual(self.hb["2015-01-10"], "Custom Holiday; New Holiday")
        self.assertEqual(self.hb["2015-01-11"], "Custom Holiday; New Holiday")


class TestStr(unittest.TestCase):
    def test_country(self):
        hb = CountryStub1()

        self.assertEqual(
            str(hb),
            "{'country': CS1, 'expand': True, 'language': None, "
            "'market': None, 'observed': True, 'subdiv': None, "
            "'years': set()}",
        )

        hb._populate(2013)
        self.assertEqual(
            str(hb),
            '{datetime.date(2013, 1, 1): "New Year\'s Day", '
            "datetime.date(2013, 6, 19): 'Juneteenth National Independence Day', "
            "datetime.date(2013, 7, 4): 'Independence Day', "
            "datetime.date(2013, 11, 28): 'Thanksgiving', "
            "datetime.date(2013, 12, 25): 'Christmas Day'}",
        )

    def test_market(self):
        self.assertEqual(
            str(MarketStub1()),
            "{'country': None, 'expand': True, 'language': None, "
            "'market': MS1, 'observed': True, 'subdiv': None, "
            "'years': set()}",
        )


class TestSubstitutedHolidays(unittest.TestCase):
    class CountryStub(CountryStub2, StaticHolidays):
        country = "HB"

        def __init__(self, cls, *args, **kwargs) -> None:
            StaticHolidays.__init__(self, cls)
            super().__init__(*args, **kwargs)

    def test_populate_substituted_holidays(self):
        hb = CountryStub1()
        self.assertTrue(hb.has_special_holidays)
        self.assertTrue(hb.has_substituted_holidays)

        hb._populate(1991)
        self.assertIn("1991-01-07", hb)
        self.assertIn("1991-01-08", hb)
        self.assertIn("12/01/1991", hb["1991-01-07"])
        self.assertIn("13/01/1991", hb["1991-01-08"])

    def test_no_substituted_date_format(self):
        class SubstitutedHolidays:
            special_public_holidays = {
                1991: (
                    (JAN, 12, JAN, 7),
                    (1991, JAN, 13, JAN, 8),
                ),
            }
            substituted_label = "From %s"

        self.assertRaises(ValueError, lambda: self.CountryStub(SubstitutedHolidays))

    def test_no_substituted_holidays(self):
        class EmptySubstitutedHolidays:
            special_public_holidays = {}
            substituted_date_format = "%d/%m/%Y"
            substituted_label = "From %s"

        class NoSubstitutedHolidays:
            substituted_date_format = "%d/%m/%Y"
            substituted_label = "From %s"

        for cls in (EmptySubstitutedHolidays, NoSubstitutedHolidays):
            hb = self.CountryStub(cls=cls)
            self.assertFalse(hb.has_special_holidays)
            self.assertFalse(hb.has_substituted_holidays)

            hb._populate(1991)
            self.assertNotIn("1991-01-07", hb)
            self.assertNotIn("1991-01-08", hb)

    def test_no_substituted_label(self):
        class SubstitutedHolidays:
            special_public_holidays = {
                1991: (
                    (JAN, 7, JAN, 12),
                    (JAN, 8, JAN, 13, 1991),
                ),
            }
            substituted_date_format = "%d/%m/%Y"

        self.assertRaises(ValueError, lambda: self.CountryStub(SubstitutedHolidays))


class TestWorkdays(unittest.TestCase):
    def setUp(self):
        self.hb = CountryStub6(years=2024)

    def test_is_working_day(self):
        self.assertTrue(self.hb.is_working_day("2024-02-12"))
        self.assertFalse(self.hb.is_working_day("2024-02-17"))
        self.assertFalse(self.hb.is_working_day("2024-02-19"))
        self.assertTrue(self.hb.is_working_day("2024-02-24"))

        self.assertTrue(self.hb.is_working_day("2024-04-30"))
        self.assertFalse(self.hb.is_working_day("2024-05-01"))
        self.assertFalse(self.hb.is_working_day("2024-05-02"))
        self.assertTrue(self.hb.is_working_day("2024-05-03"))

        self.assertFalse(self.hb.is_working_day("2024-10-07"))  # substituted from Oct 12 (SAT).
        self.assertFalse(self.hb.is_working_day("2024-10-12"))  # Columbus Day holiday.

    def test_get_nth_working_day(self):
        self.assertEqual(self.hb.get_nth_working_day("2024-01-04", 0), date(2024, 1, 4))
        self.assertEqual(self.hb.get_nth_working_day("2024-01-04", +1), date(2024, 1, 5))
        self.assertEqual(self.hb.get_nth_working_day("2024-01-04", +3), date(2024, 1, 9))
        self.assertEqual(self.hb.get_nth_working_day("2024-01-06", +1), date(2024, 1, 8))
        self.assertEqual(self.hb.get_nth_working_day("2024-01-26", -10), date(2024, 1, 12))
        self.assertEqual(self.hb.get_nth_working_day("2024-01-21", -1), date(2024, 1, 19))

        self.assertEqual(self.hb.get_nth_working_day("2024-02-15", +4), date(2024, 2, 22))
        self.assertEqual(self.hb.get_nth_working_day("2024-02-15", +5), date(2024, 2, 23))
        self.assertEqual(self.hb.get_nth_working_day("2024-02-15", +6), date(2024, 2, 24))
        self.assertEqual(self.hb.get_nth_working_day("2024-02-15", +7), date(2024, 2, 26))
        self.assertEqual(self.hb.get_nth_working_day("2024-02-26", -7), date(2024, 2, 15))
        self.assertEqual(self.hb.get_nth_working_day("2024-02-25", -7), date(2024, 2, 15))

        self.assertEqual(self.hb.get_nth_working_day("2024-04-29", +1), date(2024, 4, 30))
        self.assertEqual(self.hb.get_nth_working_day("2024-04-29", +2), date(2024, 5, 3))
        self.assertEqual(self.hb.get_nth_working_day("2024-04-29", +3), date(2024, 5, 6))
        self.assertEqual(self.hb.get_nth_working_day("2024-04-29", +4), date(2024, 5, 7))
        self.assertEqual(self.hb.get_nth_working_day("2024-05-10", -10), date(2024, 4, 24))
        self.assertEqual(self.hb.get_nth_working_day("2024-05-10", -7), date(2024, 4, 29))
        self.assertEqual(self.hb.get_nth_working_day("2024-05-10", -5), date(2024, 5, 3))

        self.assertEqual(self.hb.get_nth_working_day("2024-07-27", 0), date(2024, 7, 29))
        self.assertEqual(self.hb.get_nth_working_day("2024-07-27", 1), date(2024, 7, 29))
        self.assertEqual(self.hb.get_nth_working_day("2024-07-29", 0), date(2024, 7, 29))

    def test_get_working_days_count(self):
        self.assertEqual(self.hb.get_working_days_count("2024-01-03", "2024-01-23"), 15)
        self.assertEqual(self.hb.get_working_days_count("2024-01-23", "2024-01-03"), 15)
        self.assertEqual(self.hb.get_working_days_count("2024-01-06", "2024-01-07"), 0)
        self.assertEqual(self.hb.get_working_days_count("2024-01-16", "2024-01-16"), 1)

        self.assertEqual(self.hb.get_working_days_count("2024-02-08", "2024-02-15"), 6)
        self.assertEqual(self.hb.get_working_days_count("2024-02-15", "2024-02-22"), 5)
        self.assertEqual(self.hb.get_working_days_count("2024-02-22", "2024-02-29"), 7)

        self.assertEqual(self.hb.get_working_days_count("2024-04-29", "2024-05-03"), 3)
        self.assertEqual(self.hb.get_working_days_count("2024-04-29", "2024-05-04"), 3)
        self.assertEqual(self.hb.get_working_days_count("2024-04-29", "2024-05-05"), 3)
        self.assertEqual(self.hb.get_working_days_count("2024-04-29", "2024-05-06"), 4)


class TestClosestHoliday(unittest.TestCase):
    def setUp(self):
        self.current_year = datetime.now().year
        self.next_year = self.current_year + 1
        self.previous_year = self.current_year - 1
        self.hb = CountryStub3(years=self.current_year)
        self.next_labor_day_year = (
            self.current_year
            if datetime.now().date() <= self.hb.get_named("Custom May 1st Holiday")[0]
            else self.next_year
        )
        self.previous_labor_day_year = (
            self.current_year
            if datetime.now().date() >= self.hb.get_named("Custom May 2nd Holiday")[0]
            else self.previous_year
        )

    def test_get_closest_holiday_forward(self):
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-01-01"),
            (date(self.current_year, 5, 1), "Custom May 1st Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-04-30"),
            (date(self.current_year, 5, 1), "Custom May 1st Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-05-01"),
            (date(self.current_year, 5, 2), "Custom May 2nd Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-05-02"),
            (date(self.next_year, 5, 1), "Custom May 1st Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.next_year}-01-01"),
            (date(self.next_year, 5, 1), "Custom May 1st Holiday"),
        )

        self.assertIn(
            self.hb.get_closest_holiday(),
            [
                (date(self.next_labor_day_year, 5, 1), "Custom May 1st Holiday"),
                (date(self.next_labor_day_year, 5, 2), "Custom May 2nd Holiday"),
            ],
        )

    def test_get_closest_holiday_backward(self):
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-12-31", direction="backward"),
            (date(self.current_year, 5, 2), "Custom May 2nd Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-05-02", direction="backward"),
            (date(self.current_year, 5, 1), "Custom May 1st Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.current_year}-04-30", direction="backward"),
            (date(self.previous_year, 5, 2), "Custom May 2nd Holiday"),
        )
        self.assertEqual(
            self.hb.get_closest_holiday(f"{self.previous_year}-12-31", direction="backward"),
            (date(self.previous_year, 5, 2), "Custom May 2nd Holiday"),
        )

        self.assertIn(
            self.hb.get_closest_holiday(direction="backward"),
            [
                (date(self.previous_labor_day_year, 5, 2), "Custom May 2nd Holiday"),
                (date(self.current_year, 5, 1), "Custom May 1st Holiday"),
            ],
        )

    def test_get_closest_holiday_corner_cases(self):
        us = US()
        # check for date before start of calendar
        self.assertIsNone(us.get_closest_holiday("1777-01-01", direction="backward"))

        # check for date after end of calendar
        self.assertIsNone(us.get_closest_holiday("2100-12-31"))

    def test_get_closest_holiday_after_empty_year(self):
        ua = UA(years=2025)
        # check for date if a year has no holidays
        ua._add_holiday_jan_1("Custom holiday")
        self.assertEqual(
            ua.get_closest_holiday("2022-03-08"), (date(2025, 1, 1), "Custom holiday")
        )

    def test_get_closest_holiday_unsorted_calendars(self):
        us_calendar = US(years=2024)

        self.assertEqual(
            us_calendar.get_closest_holiday(date(2024, 2, 1)),
            (date(2024, 2, 19), "Washington's Birthday"),
        )

        # check for date before start of calendar
        self.assertEqual(
            us_calendar.get_closest_holiday(date(2024, 2, 1), direction="backward"),
            (date(2024, 1, 15), "Martin Luther King Jr. Day"),
        )

    def test_get_closest_holiday_invalid_direction(self):
        self.assertRaises(
            AttributeError, lambda: HolidayBase().get_closest_holiday(direction="invalid")
        )



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from unittest import TestCase

from holidays.countries.anguilla import Anguilla
from tests.common import CommonCountryTests


class TestAnguilla(CommonCountryTests, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass(Anguilla)

    def test_special_holiday(self):
        self.assertHolidayName("Royal Wedding of Prince William & Kate Middleton", "2011-04-29")
        self.assertHolidayName(
            "Diamond Jubilee Celebration of Her Majesty The Queen", "2012-06-04"
        )
        self.assertHolidayName(
            "Mourning the Death of Her Majesty The Queen Elizabeth II", "2022-09-19"
        )
        self.assertHolidayName("Special Public Holiday", "2025-02-28")

    def test_new_years_day(self):
        name = "New Year's Day"
        self.assertHolidayName(name, (f"{year}-01-01" for year in self.full_range))
        obs_dts = (
            "2011-01-03",
            "2012-01-02",
            "2017-01-02",
            "2022-01-03",
            "2023-01-02",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_james_ronald_webster_day(self):
        name = "James Ronald Webster Day"
        self.assertHolidayName(name, (f"{year}-03-02" for year in range(2010, self.end_year)))
        self.assertNoHolidayName(name, range(self.start_year, 2010))
        obs_dts = (
            "2013-03-04",
            "2014-03-03",
            "2019-03-04",
            "2024-03-04",
            "2025-03-03",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_good_friday(self):
        name = "Good Friday"
        self.assertHolidayName(
            name,
            "2020-04-10",
            "2021-04-02",
            "2022-04-15",
            "2023-04-07",
            "2024-03-29",
            "2025-04-18",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_monday(self):
        name = "Easter Monday"
        self.assertHolidayName(
            name,
            "2020-04-13",
            "2021-04-05",
            "2022-04-18",
            "2023-04-10",
            "2024-04-01",
            "2025-04-21",
        )
        self.assertHolidayName(name, self.full_range)

    def test_easter_sunday(self):
        name = "Easter Sunday"
        self.assertHolidayName(
            name,
            "2020-04-12",
            "2021-04-04",
            "2022-04-17",
            "2023-04-09",
            "2024-03-31",
            "2025-04-20",
        )
        self.assertHolidayName(name, self.full_range)

    def test_labor_day(self):
        name = "Labour Day"
        self.assertHolidayName(name, (f"{year}-05-01" for year in self.full_range))
        obs_dts = (
            "2011-05-02",
            "2016-05-02",
            "2021-05-03",
            "2022-05-02",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_whit_monday(self):
        name = "Whit Monday"
        self.assertHolidayName(
            name,
            "2020-06-01",
            "2021-05-24",
            "2022-06-06",
            "2023-05-29",
            "2024-05-20",
            "2025-06-09",
        )
        self.assertHolidayName(name, self.full_range)

    def test_queen_birthday(self):
        name = "Celebration of the Birthday of Her Majesty the Queen"
        self.assertHolidayName(
            name,
            "2016-06-13",
            "2017-06-12",
            "2018-06-11",
            "2019-06-10",
            "2020-06-08",
            "2021-06-14",
            "2022-06-03",
        )
        self.assertHolidayName(name, range(self.start_year, 2023))
        self.assertNoHolidayName(name, range(2023, self.end_year))

    def test_king_birthday(self):
        name = "Celebration of the Birthday of His Majesty the King"
        self.assertHolidayName(
            name,
            "2023-06-19",
            "2024-06-17",
            "2025-06-16",
        )
        self.assertHolidayName(name, range(2023, self.end_year))
        self.assertNoHolidayName(name, range(self.start_year, 2023))

    def test_anguilla_day(self):
        name = "Anguilla Day"
        self.assertHolidayName(name, (f"{year}-05-30" for year in self.full_range))
        obs_dts = (
            "2004-06-01",
            "2009-06-02",
            "2010-05-31",
            "2015-06-01",
            "2020-06-02",
            "2021-05-31",
            "2039-05-31",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_august_monday(self):
        name = "August Monday"
        self.assertHolidayName(
            name,
            "2020-08-03",
            "2021-08-02",
            "2022-08-01",
            "2023-08-07",
            "2024-08-05",
            "2025-08-04",
        )
        self.assertHolidayName(name, self.full_range)

    def test_august_thursday(self):
        name = "August Thursday"
        self.assertHolidayName(
            name,
            "2020-08-06",
            "2021-08-05",
            "2022-08-04",
            "2023-08-10",
            "2024-08-08",
            "2025-08-07",
        )
        self.assertHolidayName(name, self.full_range)

    def test_constitution_day(self):
        name = "Constitution Day"
        self.assertHolidayName(
            name,
            "2020-08-07",
            "2021-08-06",
            "2022-08-05",
            "2023-08-11",
            "2024-08-09",
            "2025-08-08",
        )
        self.assertHolidayName(name, self.full_range)

    def test_national_heroes_and_heroines_day(self):
        name_2001 = "Separation Day"
        self.assertHolidayName(
            name_2001, (f"{year}-12-19" for year in range(self.start_year, 2011))
        )
        self.assertNoHolidayName(name_2001, range(2011, self.end_year))
        obs_dts_2001 = (
            "2004-12-17",
            "2009-12-18",
            "2010-12-17",
        )
        self.assertHolidayName(f"{name_2001} (observed)", obs_dts_2001)
        self.assertNoNonObservedHoliday(obs_dts_2001)

        name_2011 = "National Heroes and Heroines Day"
        self.assertHolidayName(name_2011, (f"{year}-12-19" for year in range(2011, self.end_year)))
        self.assertNoHolidayName(name_2011, range(self.start_year, 2011))
        obs_dts_2011 = (
            "2015-12-18",
            "2020-12-18",
            "2021-12-17",
        )
        self.assertHolidayName(f"{name_2011} (observed)", obs_dts_2011)
        self.assertNoNonObservedHoliday(obs_dts_2011)

    def test_christmas_day(self):
        name = "Christmas Day"
        self.assertHolidayName(name, (f"{year}-12-25" for year in self.full_range))
        obs_dts = (
            "2010-12-27",
            "2011-12-27",
            "2016-12-27",
            "2021-12-27",
            "2022-12-27",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_boxing_day(self):
        name = "Boxing Day"
        self.assertHolidayName(name, (f"{year}-12-26" for year in self.full_range))
        obs_dts = (
            "2009-12-28",
            "2010-12-28",
            "2015-12-28",
            "2020-12-28",
            "2021-12-28",
        )
        self.assertHolidayName(f"{name} (observed)", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_l10n_default(self):
        self.assertLocalizedHolidays(
            ("2022-01-01", "New Year's Day"),
            ("2022-01-03", "New Year's Day (observed)"),
            ("2022-03-02", "James Ronald Webster Day"),
            ("2022-04-15", "Good Friday"),
            ("2022-04-17", "Easter Sunday"),
            ("2022-04-18", "Easter Monday"),
            ("2022-05-01", "Labour Day"),
            ("2022-05-02", "Labour Day (observed)"),
            ("2022-05-30", "Anguilla Day"),
            ("2022-06-03", "Celebration of the Birthday of Her Majesty the Queen"),
            ("2022-06-06", "Whit Monday"),
            ("2022-08-01", "August Monday"),
            ("2022-08-04", "August Thursday"),
            ("2022-08-05", "Constitution Day"),
            ("2022-09-19", "Mourning the Death of Her Majesty The Queen Elizabeth II"),
            ("2022-12-19", "National Heroes and Heroines Day"),
            ("2022-12-25", "Christmas Day"),
            ("2022-12-26", "Boxing Day"),
            ("2022-12-27", "Christmas Day (observed)"),
        )

    def test_l10n_en_us(self):
        self.assertLocalizedHolidays(
            "en_US",
            ("2022-01-01", "New Year's Day"),
            ("2022-01-03", "New Year's Day (observed)"),
            ("2022-03-02", "James Ronald Webster Day"),
            ("2022-04-15", "Good Friday"),
            ("2022-04-17", "Easter Sunday"),
            ("2022-04-18", "Easter Monday"),
            ("2022-05-01", "Labor Day"),
            ("2022-05-02", "Labor Day (observed)"),
            ("2022-05-30", "Anguilla Day"),
            ("2022-06-03", "Queen's Birthday"),
            ("2022-06-06", "Whit Monday"),
            ("2022-08-01", "August Monday"),
            ("2022-08-04", "August Thursday"),
            ("2022-08-05", "Constitution Day"),
            ("2022-09-19", "Mourning the Death of Her Majesty The Queen Elizabeth II"),
            ("2022-12-19", "National Heroes and Heroines Day"),
            ("2022-12-25", "Christmas Day"),
            ("2022-12-26", "Boxing Day"),
            ("2022-12-27", "Christmas Day (observed)"),
        )



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from unittest import TestCase

from holidays.countries.laos import Laos
from tests.common import CommonCountryTests


class TestLaos(CommonCountryTests, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass(Laos)

    def test_special_bank_holidays(self):
        self.assertBankHoliday("2015-01-02")
        self.assertNoBankNonObservedHoliday(
            "2012-10-08",
            "2017-10-09",
            "2018-10-08",
            "2023-10-09",
        )

    def test_special_public_holidays(self):
        dts = ("2015-04-17",)
        obs_dts = (
            "2011-04-13",
            "2020-04-17",
        )
        self.assertHoliday(dts, obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_special_workdays(self):
        self.assertWorkdayHoliday("2019-07-22")
        self.assertNoWorkdayNonObservedHoliday("2019-07-22")

    def test_establishment_day_of_the_bol(self):
        name = "ວັນສ້າງຕັ້ງທະນາຄານແຫ່ງ ສປປ ລາວ"
        self.assertNoHolidayName(name)
        self.assertBankHolidayName(name, (f"{year}-10-07" for year in self.full_range))

        obs_dts = (
            "2012-10-08",
            "2017-10-09",
            "2018-10-08",
            "2023-10-09",
        )
        self.assertBankHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoBankNonObservedHoliday(obs_dts)

    def test_lao_year_end_bank_holiday(self):
        name = "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"
        self.assertNoHolidayName(name)
        self.assertBankHolidayName(
            name,
            "2020-12-29",
            "2020-12-30",
            "2020-12-31",
            "2021-12-29",
            "2021-12-30",
            "2021-12-31",
            "2022-12-28",
            "2022-12-29",
            "2022-12-30",
            "2023-12-27",
            "2023-12-28",
            "2023-12-29",
            "2024-12-27",
            "2024-12-30",
            "2024-12-31",
            "2025-12-29",
            "2025-12-30",
            "2025-12-31",
        )
        self.assertBankHolidayNameCount(name, 3, self.full_range)

    def test_new_years_day(self):
        name = "ວັນປີໃໝ່ສາກົນ"
        self.assertHolidayName(name, (f"{year}-01-01" for year in self.full_range))

        obs_dts = (
            "2012-01-02",
            "2017-01-02",
            "2022-01-03",
            "2023-01-02",
        )
        self.assertHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_international_women_rights_day(self):
        name = "ວັນແມ່ຍິງສາກົນ"
        self.assertHolidayName(name, (f"{year}-03-08" for year in self.full_range))

        obs_dts = (
            "2015-03-09",
            "2020-03-09",
        )
        self.assertHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_laos_new_year_day(self):
        name = "ບຸນປີໃໝ່ລາວ"
        songkran_years_apr_13_15 = {2012, 2017}
        songkran_years_apr_13_16 = {2016, 2020, 2024}
        for year in self.full_range:
            if year in songkran_years_apr_13_15:
                self.assertHolidayName(name, f"{year}-04-13", f"{year}-04-14", f"{year}-04-15")
            elif year in songkran_years_apr_13_16:
                self.assertHolidayName(
                    name,
                    f"{year}-04-13",
                    f"{year}-04-14",
                    f"{year}-04-15",
                    f"{year}-04-16",
                )
            else:
                self.assertHolidayName(name, f"{year}-04-14", f"{year}-04-15", f"{year}-04-16")

        obs_dts = (
            "2012-04-16",
            "2012-04-17",
            "2013-04-17",
            "2016-04-18",
            "2017-04-17",
            "2018-04-17",
            "2018-04-18",
            "2019-04-17",
            "2022-04-18",
            "2023-04-17",
            "2023-04-18",
            "2024-04-17",
            "2024-04-18",
        )
        self.assertHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_international_labor_day(self):
        name = "ວັນກຳມະກອນສາກົນ"
        self.assertHolidayName(name, (f"{year}-05-01" for year in self.full_range))

        obs_dts = (
            "2016-05-02",
            "2021-05-03",
            "2022-05-02",
        )
        self.assertHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_international_childrens_day(self):
        name = "ວັນເດັກສາກົນ"
        # PUBLIC.
        self.assertHolidayName(name, (f"{year}-06-01" for year in range(1990, 2018)))
        self.assertNoHolidayName(name, range(self.start_year, 1990), range(2018, self.end_year))
        # WORKDAY.
        self.assertWorkdayHolidayName(
            name, (f"{year}-06-01" for year in range(2018, self.end_year))
        )
        self.assertNoWorkdayHolidayName(name, range(self.start_year, 2018))

    def test_lao_national_day(self):
        name = "ວັນຊາດ"
        self.assertHolidayName(name, (f"{year}-12-02" for year in self.full_range))

        obs_dts = (
            "2012-12-03",
            "2017-12-04",
            "2018-12-03",
            "2023-12-04",
        )
        self.assertHolidayName(f"ພັກຊົດເຊີຍ{name}", obs_dts)
        self.assertNoNonObservedHoliday(obs_dts)

    def test_makha_bousa_festival(self):
        name = "ວັນບຸນມາຂະບູຊາ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-02-08",
            "2021-01-28",
            "2022-02-16",
            "2023-02-05",
            "2024-02-24",
            "2025-02-12",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_visakha_bousa_festival(self):
        name = "ວັນບຸນວິສາຂະບູຊາ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-05-06",
            "2021-04-26",
            "2022-05-15",
            "2023-05-04",
            "2024-05-22",
            "2025-05-11",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_khao_phansa_begin_of_buddhist_lent(self):
        name = "ວັນບຸນເຂົ້າພັນສາ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-07-05",
            "2021-07-24",
            "2022-07-13",
            "2023-08-01",
            "2024-07-20",
            "2025-07-10",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_haw_khao_padapdin_rice_growing_festival(self):
        name = "ວັນບຸນຫໍ່ເຂົ້າປະດັບດິນ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-08-18",
            "2021-09-06",
            "2022-08-26",
            "2023-09-14",
            "2024-09-02",
            "2025-08-23",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_haw_khao_salark_ancestor_festival(self):
        name = "ວັນບຸນຫໍ່ເຂົ້າສະຫຼາກ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-09-02",
            "2021-09-21",
            "2022-09-10",
            "2023-09-29",
            "2024-09-17",
            "2025-09-07",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_awk_phansa_end_of_buddhist_lent(self):
        name = "ວັນບຸນອອກພັນສາ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-10-02",
            "2021-10-21",
            "2022-10-10",
            "2023-10-29",
            "2024-10-17",
            "2025-10-07",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_suang_huea_vientiane_boat_racing_festival(self):
        name = "ວັນບຸນຊ່ວງເຮືອ ນະຄອນຫຼວງວຽງຈັນ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-10-03",
            "2021-10-22",
            "2022-10-11",
            "2023-10-30",
            "2024-10-18",
            "2025-10-08",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_boun_that_luang_festival(self):
        name = "ວັນບຸນທາດຫລວງ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name,
            "2020-10-31",
            "2021-11-19",
            "2022-11-08",
            "2023-11-27",
            "2024-11-15",
            "2025-11-05",
        )
        self.assertSchoolHolidayName(name, self.full_range)

    def test_national_teacher_day(self):
        name = "ວັນຄູແຫ່ງຊາດ"
        self.assertNoHolidayName(name)
        self.assertSchoolHolidayName(
            name, (f"{year}-10-07" for year in range(1994, self.end_year))
        )
        self.assertNoSchoolHolidayName(name, range(self.start_year, 1994))

    def test_lao_peoples_armed_force_day(self):
        name = "ວັນສ້າງຕັ້ງກອງທັບປະຊາຊົນລາວ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-01-20" for year in self.full_range))

    def test_lao_federation_of_trade_unions_day(self):
        name = "ວັນສ້າງຕັ້ງສະຫະພັນກໍາມະບານລາວ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-02-01" for year in self.full_range))

    def test_establishment_day_of_the_lao_peoples_revolutionary_party(self):
        name = "ວັນສ້າງຕັ້ງພັກປະຊາຊົນປະຕິວັດລາວ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-03-22" for year in self.full_range))

    def test_lao_peoples_revolutionary_youth_union_day(self):
        name = "ວັນສ້າງຕັ້ງສູນກາງຊາວໜຸ່ມປະຊາຊົນປະຕິວັດລາວ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-04-14" for year in self.full_range))

    def test_national_arbor_day(self):
        name = "ວັນປູກຕົ້ນໄມ້ແຫ່ງຊາດ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(
            name, (f"{year}-06-01" for year in range(1989, self.end_year))
        )
        self.assertNoWorkdayHolidayName(name, range(self.start_year, 1989))

    def test_president_souphanouvongs_birthday(self):
        name = "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ສຸພານຸວົງ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-07-13" for year in self.full_range))

    def test_the_national_day_for_wildlife_and_aquatic_animal_conservation(self):
        name = "ວັນປ່ອຍປາ ແລະ ວັນອະນຸລັກສັດນ້ຳ-ສັດປ່າແຫ່ງຊາດ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(
            name, (f"{year}-07-13" for year in range(1997, self.end_year))
        )
        self.assertNoWorkdayHolidayName(name, range(self.start_year, 1997))

    def test_establishment_day_of_the_lao_womens_union(self):
        name = "ວັນສ້າງຕັ້ງສະຫະພັນແມ່ຍິງລາວ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-07-20" for year in self.full_range))

    def test_lao_national_mass_media_and_publishing_day(self):
        name = "ວັນສື່ມວນຊົນແຫ່ງຊາດ ແລະ ວັນພິມຈໍາໜ່າຍ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-08-13" for year in self.full_range))

    def test_lao_national_constitution_day(self):
        name = "ວັນລັດຖະທໍາມະນູນແຫ່ງຊາດ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(
            name, (f"{year}-08-15" for year in range(1991, self.end_year))
        )
        self.assertNoWorkdayHolidayName(name, range(self.start_year, 1991))

    def test_national_uprising_day(self):
        name = "ວັນຍຶດອຳນາດທົ່ວປະເທດ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-08-23" for year in self.full_range))

    def test_independence_declaration_day(self):
        name = "ວັນປະກາດເອກະລາດ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(name, (f"{year}-10-12" for year in self.full_range))

    def test_president_kaysone_promvihanes_birthday(self):
        name = "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ໄກສອນ ພົມວິຫານ"
        self.assertNoHolidayName(name)
        self.assertWorkdayHolidayName(
            name, (f"{year}-12-13" for year in range(1991, self.end_year))
        )
        self.assertNoWorkdayHolidayName(name, range(self.start_year, 1991))

    def test_2023_public_holiday(self):
        self.assertHolidaysInYear(
            2023,
            ("2023-01-01", "ວັນປີໃໝ່ສາກົນ"),
            ("2023-01-02", "ພັກຊົດເຊີຍວັນປີໃໝ່ສາກົນ"),
            ("2023-03-08", "ວັນແມ່ຍິງສາກົນ"),
            ("2023-04-14", "ບຸນປີໃໝ່ລາວ"),
            ("2023-04-15", "ບຸນປີໃໝ່ລາວ"),
            ("2023-04-16", "ບຸນປີໃໝ່ລາວ"),
            ("2023-04-17", "ພັກຊົດເຊີຍບຸນປີໃໝ່ລາວ"),
            ("2023-04-18", "ພັກຊົດເຊີຍບຸນປີໃໝ່ລາວ"),
            ("2023-05-01", "ວັນກຳມະກອນສາກົນ"),
            ("2023-12-02", "ວັນຊາດ"),
            ("2023-12-04", "ພັກຊົດເຊີຍວັນຊາດ"),
        )

    def test_2023_bank_holidays(self):
        # Dec 31 is Sunday.
        self.assertBankHolidaysInYear(
            2023,
            ("2023-10-07", "ວັນສ້າງຕັ້ງທະນາຄານແຫ່ງ ສປປ ລາວ"),
            ("2023-10-09", "ພັກຊົດເຊີຍວັນສ້າງຕັ້ງທະນາຄານແຫ່ງ ສປປ ລາວ"),
            ("2023-12-27", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
            ("2023-12-28", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
            ("2023-12-29", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
        )

    def test_2023_school_holidays(self):
        self.assertSchoolHolidaysInYear(
            2023,
            ("2023-02-05", "ວັນບຸນມາຂະບູຊາ"),
            ("2023-05-04", "ວັນບຸນວິສາຂະບູຊາ"),
            ("2023-08-01", "ວັນບຸນເຂົ້າພັນສາ"),
            ("2023-09-14", "ວັນບຸນຫໍ່ເຂົ້າປະດັບດິນ"),
            ("2023-09-29", "ວັນບຸນຫໍ່ເຂົ້າສະຫຼາກ"),
            ("2023-10-07", "ວັນຄູແຫ່ງຊາດ"),
            ("2023-10-29", "ວັນບຸນອອກພັນສາ"),
            ("2023-10-30", "ວັນບຸນຊ່ວງເຮືອ ນະຄອນຫຼວງວຽງຈັນ"),
            ("2023-11-27", "ວັນບຸນທາດຫລວງ"),
        )

    def test_2023_workdays(self):
        self.assertWorkdayHolidaysInYear(
            2023,
            ("2023-01-20", "ວັນສ້າງຕັ້ງກອງທັບປະຊາຊົນລາວ"),
            ("2023-02-01", "ວັນສ້າງຕັ້ງສະຫະພັນກໍາມະບານລາວ"),
            ("2023-03-22", "ວັນສ້າງຕັ້ງພັກປະຊາຊົນປະຕິວັດລາວ"),
            ("2023-04-14", "ວັນສ້າງຕັ້ງສູນກາງຊາວໜຸ່ມປະຊາຊົນປະຕິວັດລາວ"),
            ("2023-06-01", "ວັນປູກຕົ້ນໄມ້ແຫ່ງຊາດ; ວັນເດັກສາກົນ"),
            ("2023-07-13", "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ສຸພານຸວົງ; ວັນປ່ອຍປາ ແລະ ວັນອະນຸລັກສັດນ້ຳ-ສັດປ່າແຫ່ງຊາດ"),
            ("2023-07-20", "ວັນສ້າງຕັ້ງສະຫະພັນແມ່ຍິງລາວ"),
            ("2023-08-13", "ວັນສື່ມວນຊົນແຫ່ງຊາດ ແລະ ວັນພິມຈໍາໜ່າຍ"),
            ("2023-08-15", "ວັນລັດຖະທໍາມະນູນແຫ່ງຊາດ"),
            ("2023-08-23", "ວັນຍຶດອຳນາດທົ່ວປະເທດ"),
            ("2023-10-12", "ວັນປະກາດເອກະລາດ"),
            ("2023-12-13", "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ໄກສອນ ພົມວິຫານ"),
        )

    def test_l10n_default(self):
        self.assertLocalizedHolidays(
            ("2022-01-01", "ວັນປີໃໝ່ສາກົນ"),
            ("2022-01-03", "ພັກຊົດເຊີຍວັນປີໃໝ່ສາກົນ"),
            ("2022-01-20", "ວັນສ້າງຕັ້ງກອງທັບປະຊາຊົນລາວ"),
            ("2022-02-01", "ວັນສ້າງຕັ້ງສະຫະພັນກໍາມະບານລາວ"),
            ("2022-02-16", "ວັນບຸນມາຂະບູຊາ"),
            ("2022-03-08", "ວັນແມ່ຍິງສາກົນ"),
            ("2022-03-22", "ວັນສ້າງຕັ້ງພັກປະຊາຊົນປະຕິວັດລາວ"),
            ("2022-04-14", "ບຸນປີໃໝ່ລາວ; ວັນສ້າງຕັ້ງສູນກາງຊາວໜຸ່ມປະຊາຊົນປະຕິວັດລາວ"),
            ("2022-04-15", "ບຸນປີໃໝ່ລາວ"),
            ("2022-04-16", "ບຸນປີໃໝ່ລາວ"),
            ("2022-04-18", "ພັກຊົດເຊີຍບຸນປີໃໝ່ລາວ"),
            ("2022-05-01", "ວັນກຳມະກອນສາກົນ"),
            ("2022-05-02", "ພັກຊົດເຊີຍວັນກຳມະກອນສາກົນ"),
            ("2022-05-15", "ວັນບຸນວິສາຂະບູຊາ"),
            ("2022-06-01", "ວັນປູກຕົ້ນໄມ້ແຫ່ງຊາດ; ວັນເດັກສາກົນ"),
            (
                "2022-07-13",
                "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ສຸພານຸວົງ; ວັນບຸນເຂົ້າພັນສາ; ວັນປ່ອຍປາ ແລະ ວັນອະນຸລັກສັດນ້ຳ-ສັດປ່າແຫ່ງຊາດ",
            ),
            ("2022-07-20", "ວັນສ້າງຕັ້ງສະຫະພັນແມ່ຍິງລາວ"),
            ("2022-08-13", "ວັນສື່ມວນຊົນແຫ່ງຊາດ ແລະ ວັນພິມຈໍາໜ່າຍ"),
            ("2022-08-15", "ວັນລັດຖະທໍາມະນູນແຫ່ງຊາດ"),
            ("2022-08-23", "ວັນຍຶດອຳນາດທົ່ວປະເທດ"),
            ("2022-08-26", "ວັນບຸນຫໍ່ເຂົ້າປະດັບດິນ"),
            ("2022-09-10", "ວັນບຸນຫໍ່ເຂົ້າສະຫຼາກ"),
            ("2022-10-07", "ວັນຄູແຫ່ງຊາດ; ວັນສ້າງຕັ້ງທະນາຄານແຫ່ງ ສປປ ລາວ"),
            ("2022-10-10", "ວັນບຸນອອກພັນສາ"),
            ("2022-10-11", "ວັນບຸນຊ່ວງເຮືອ ນະຄອນຫຼວງວຽງຈັນ"),
            ("2022-10-12", "ວັນປະກາດເອກະລາດ"),
            ("2022-11-08", "ວັນບຸນທາດຫລວງ"),
            ("2022-12-02", "ວັນຊາດ"),
            ("2022-12-13", "ວັນຄ້າຍວັນເກີດ ທ່ານ ປະທານ ໄກສອນ ພົມວິຫານ"),
            ("2022-12-28", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
            ("2022-12-29", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
            ("2022-12-30", "ສາມວັນລັດຖະການສຸດທ້າຍຂອງທຸກໆປີ"),
        )

    def test_l10n_en_us(self):
        self.assertLocalizedHolidays(
            "en_US",
            ("2022-01-01", "New Year's Day"),
            ("2022-01-03", "New Year's Day (in lieu)"),
            ("2022-01-20", "Lao People's Armed Force Day"),
            ("2022-02-01", "Lao Federation of Trade Union's Day"),
            ("2022-02-16", "Makha Bousa Festival"),
            ("2022-03-08", "International Women's Rights Day"),
            ("2022-03-22", "Establishment Day of the Lao People's Revolutionary Party"),
            ("2022-04-14", "Lao New Year's Day; Lao People's Revolutionary Youth Union Day"),
            ("2022-04-15", "Lao New Year's Day"),
            ("2022-04-16", "Lao New Year's Day"),
            ("2022-04-18", "Lao New Year's Day (in lieu)"),
            ("2022-05-01", "International Labor Day"),
            ("2022-05-02", "International Labor Day (in lieu)"),
            ("2022-05-15", "Visakha Bousa Festival"),
            ("2022-06-01", "International Children's Day; National Arbor Day"),
            (
                "2022-07-13",
                "Begin of Buddhist Lent; President Souphanouvong's Birthday; "
                "The National Day for Wildlife and Aquatic Animal Conservation",
            ),
            ("2022-07-20", "Establishment Day of the Lao Women's Union"),
            ("2022-08-13", "Lao National Mass Media and Publishing Day"),
            ("2022-08-15", "Lao National Constitution Day"),
            ("2022-08-23", "National Uprising Day"),
            ("2022-08-26", "Boun Haw Khao Padapdin"),
            ("2022-09-10", "Boun Haw Khao Salark"),
            ("2022-10-07", "Establishment Day of the BOL; National Teacher Day"),
            ("2022-10-10", "End of Buddhist Lent"),
            ("2022-10-11", "Vientiane Boat Racing Festival"),
            ("2022-10-12", "Independence Declaration Day"),
            ("2022-11-08", "Boun That Luang Festival"),
            ("2022-12-02", "Lao National Day"),
            ("2022-12-13", "President Kaysone Phomvihane's Birthday"),
            ("2022-12-28", "Lao Year-End Bank Holiday"),
            ("2022-12-29", "Lao Year-End Bank Holiday"),
            ("2022-12-30", "Lao Year-End Bank Holiday"),
        )

    def test_l10n_th(self):
        self.assertLocalizedHolidays(
            "th",
            ("2022-01-01", "วันปีใหม่สากล"),
            ("2022-01-03", "ชดเชยวันปีใหม่สากล"),
            ("2022-01-20", "วันก่อตั้งกองทัพประชาชนลาว"),
            ("2022-02-01", "วันก่อตั้งสหพันธ์กำมะบานลาว"),
            ("2022-02-16", "วันมาฆบูชา"),
            ("2022-03-08", "วันสตรีสากล"),
            ("2022-03-22", "วันก่อตั้งพรรคประชาชนปฏิวัติลาว"),
            ("2022-04-14", "วันก่อตั้งศูนย์ซาวหนุ่มประชาชนปฏิวัติลาว; วันปีใหม่ลาว"),
            ("2022-04-15", "วันปีใหม่ลาว"),
            ("2022-04-16", "วันปีใหม่ลาว"),
            ("2022-04-18", "ชดเชยวันปีใหม่ลาว"),
            ("2022-05-01", "วันแรงงานสากล"),
            ("2022-05-02", "ชดเชยวันแรงงานสากล"),
            ("2022-05-15", "วันวิสาขบูชา"),
            ("2022-06-01", "วันปลูกต้นไม้แห่งชาติ; วันเด็กสากล"),
            (
                "2022-07-13",
                "วันคล้ายวันเกิดท่านประธานสุภานุวงศ์; วันอนุรักษ์สัตว์น้ำ สัตว์ป่า และวันปล่อยปลาแห่งชาติ; วันเข้าพรรษา",
            ),
            ("2022-07-20", "วันก่อตั้งสหภาพแม่หญิงลาว"),
            ("2022-08-13", "วันสื่อสารมวลชนและการพิมพ์แห่งชาติ"),
            ("2022-08-15", "วันรัฐธรรมนูญแห่งชาติ"),
            ("2022-08-23", "วันยึดอำนาจทั่วประเทศ"),
            ("2022-08-26", "วันบุญข้าวประดับดิน"),
            ("2022-09-10", "วันข้าวบุญข้าวสาก"),
            ("2022-10-07", "วันก่อตั้งธนาคารแห่ง สปป. ลาว; วันครูแห่งชาติ"),
            ("2022-10-10", "วันออกพรรษา"),
            ("2022-10-11", "วันงานบุญแข่งเรือ นครหลวงเวียงจันทน์"),
            ("2022-10-12", "วันประกาศเอกราช"),
            ("2022-11-08", "วันงานพระธาตุหลวง"),
            ("2022-12-02", "วันชาติ สปป. ลาว"),
            ("2022-12-13", "วันคล้ายวันเกิดท่านประธานไกสอน พมวิหาน"),
            ("2022-12-28", "วันหยุดสิ้นปีของสถาบันการเงิน"),
            ("2022-12-29", "วันหยุดสิ้นปีของสถาบันการเงิน"),
            ("2022-12-30", "วันหยุดสิ้นปีของสถาบันการเงิน"),
        )
